#include "gtable.h"
#include "gtableimpl.h"

#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

/* Forward definitions */
static Gtable_entry *iGtableFind( Gtable *ctx, int row, int col );
static Gtable_entry *iGtableNew( void );
static int iGtableMessageLen( const char *fmt );

/* #define DEBUG(a) a */
#define DEBUG(a)

/* &ctx, nrow, ncol (optional) */
Gtable *GtableCreate( int nrow, int ncol )
{
    Gtable *gt;

    gt = (Gtable *)malloc( sizeof(Gtable) );
    if (!gt) return 0;

    gt->cookie  = GTABLE_COOKIE;
    gt->rows    = nrow;
    gt->cols    = ncol;
    gt->first   = 0;
    gt->cur_col = 0;
    gt->cur_row = -1;  /* Indicates no current rows */

    return gt;
}

/* ctx, row, col, string */
int GtableAddEntry( Gtable *ctx, int row, int col, const char *string )
{
    Gtable_entry *new;

    if (!GTABLE_VALID(ctx)) {
	fprintf( stderr, "Invalid gtable pointer in GtableAddEntry\n" );
	return 1;
    }

    DEBUG(printf("adding entry %d,%d\n",row,col));
    /* Find the row and column location.  Insert missing elements */
    new = iGtableFind( ctx, row, col );

    /* If there is already an entry here, do we generate an error message? */
    new->n_row = 1;
    new->n_col = 1;
    new->contents = strdup( string );
    new->est_width = -1;
    new->halign = GTABLE_DEFAULT;
    new->valign = GTABLE_VDEFAULT;
    
    return 0;
}

int GtableAddEntryFmt( Gtable *ctx, int row, int col, const char *fmt, ... )
{
    Gtable_entry *new;
    int          size;
    char         *string;
    va_list      ap;

    va_start( ap, fmt );

    if (!GTABLE_VALID(ctx)) {
	fprintf( stderr, "Invalid gtable pointer in GtableAddEntryFmt\n" );
	return 1;
    }

    DEBUG(printf("adding entry %d,%d\n",row,col));
    /* Find the row and column location.  Insert missing elements */
    new = iGtableFind( ctx, row, col );

    /* If there is already an entry here, do we generate an error message? */
    new->n_row = 1;
    new->n_col = 1;
    /* Create the new value */
    size = iGtableMessageLen( fmt );
    string = (char *)malloc( size );
    if (!string) return 1;
    string[size-1] = 0;
    vsprintf( string, fmt, ap );
    if (string[size-1]) {
	/* Error, vsprintf overran buffer */
	free( string );
	return 1;
    }
    new->contents = string;
    new->est_width = -1;
    new->halign = GTABLE_DEFAULT;
    new->valign = GTABLE_VDEFAULT;
    
    return 0;
}

int GtableAddBlock( Gtable *ctx, int row, int col, int nrow, int ncol, 
		    const char *string )
{
    if (!GTABLE_VALID(ctx)) {
	fprintf( stderr, "Invalid gtable pointer in GtableAddBlock\n" );
	return 1;
    }

    return 1;
}

int GtableOutput( Gtable *ctx, Gtable_output *method, void *extra_data )
{
    Gtable_entry *row, *col;
    int          row_num, col_num;

    if (!method) return 1;
    if (!GTABLE_VALID(ctx)) {
	fprintf( stderr, "Invalid gtable pointer in GtableOutput\n" );
	return 1;
    }

    /* Pass the external extra data to the start_table routine */
    method->start_table( ctx, method, extra_data );
    row = ctx->first;
    row_num = 0;
    while (row) {
	method->start_row( ctx, row, row_num, method->extra_data );
	col = row;
	col_num = 0;
	while (col) {
	    method->output_col( ctx, col, row_num, col_num, 
				method->extra_data );
	    col = col->next_col;
	    col_num++;
	}
	method->end_row( ctx, row, row_num, method->extra_data );
	row = row->next_row;
	row_num++;
    }
    method->end_table( ctx, method->extra_data );
    return 0;
}

int GtableAddRow( Gtable *ctx, int row, char sep_char, const char *string )
{
    const char *p = string, *p_next;
    char       *p_dup;
    int        len, col = 0;

    if (!GTABLE_VALID(ctx)) {
	fprintf( stderr, "Invalid gtable pointer in GtableAddRow\n" );
	return 1;
    }

    while (p) {
	p_next = strchr( p, sep_char );
	if (p_next) {
	    len = p_next - p;
	    p_next++;
	}
	else
	    len = strlen(p);
	p_dup = (char *)malloc( len + 1 );
	if (!p_dup) return 1;
	strncpy( p_dup, p, len );
	p_dup[len] = 0;
	if (GtableAddEntry( ctx, row, col, p_dup )) return 1;
	free( p_dup );
	col++;
	p = p_next;
    }
    return 0;
}

int GtableFree( Gtable **ctx_p )
{
    Gtable *ctx = *ctx_p;
    Gtable_entry *row, *col, *old;

    if (!GTABLE_VALID(ctx)) {
	fprintf( stderr, "Invalid gtable pointer in GtableFree\n" );
	return 1;
    }

    row = ctx->first;
    while (row) {
	col = row;
	row = row->next_row;    /* Because we free each column entry now */
	do  {
	    if (col->contents) {
		free( col->contents );
	    }
	    old = col;
	    col = col->next_col;
	    free( old );
	} while (col);
    }
    free( ctx );
    *ctx_p = 0;
    return 0;
}

int GtableNewRow( Gtable *ctx )
{
    if (!GTABLE_VALID(ctx)) {
	fprintf( stderr, "Invalid gtable pointer in GtableNewRow\n" );
	return 1;
    }

    ctx->cur_row++;
    ctx->cur_col = 0;
    return 0;
}

/* Local routines */
static Gtable_entry *iGtableFind( Gtable *ctx, int row, int col )
{
    Gtable_entry *cur;
    int          i, j;

    if (row == GTABLE_APPEND) {
	row = ctx->cur_row;
    }
    if (col == GTABLE_APPEND) {
	col = ctx->cur_col++;
    }
    DEBUG(printf("looking for %d,%d\n",row,col));
    /* Get to the correct row */
    cur = ctx->first;
    if (!cur) {
	ctx->first = cur = iGtableNew( );
    }
    for (i=0; i<row; i++) {
	if (!cur->next_row) {
	    cur->next_row = iGtableNew( );
	    if (!cur->next_row) return 0;
	}
	cur = cur->next_row;
    }
    if (ctx->rows < row) ctx->rows = row;
    /* Get to the correct column */
    for (j=0; j<col; j++) {
	if (!cur->next_col) {
	    cur->next_col = iGtableNew( );
	    if (!cur->next_col) return 0;
	}
	cur = cur->next_col;
    }
    if (ctx->cols < col) ctx->cols = col;
    return cur;
}

static Gtable_entry *iGtableNew( void )
{
    Gtable_entry *new;

    new = (Gtable_entry *)malloc( sizeof(Gtable_entry) );
    if (!new) return 0;
    new->contents = 0;
    new->n_row = 1; new->n_col = 1; 
    new->next_col = 0;
    new->next_row = 0;
    new->halign   = GTABLE_DEFAULT;
    new->valign   = GTABLE_VDEFAULT;

    return new;
}

/* Find a limit on the number of characters in a format.
   This is *very* approximate. */
static int iGtableMessageLen( const char *fmt )
{
    int size = 0;
    while (*fmt) {
	if (*fmt == '%') {
	    fmt++;
	    switch (*fmt) {
	    case '%': size++; break;
	    case 'd': size += 11; break;
	    case 'f': size += 50; break;
	    case 'x': size += 16; break;
	    case 'e': size += 50; break;
	    default:
		size += 50;
	    }
	    fmt++;
	}
	else {
	    fmt++;
	    size++;
	}
    }
    return size;
}

/*@
  GtableOutputFree - Free a gtable output method

  Input Parameter:
. output - A Gtable output method.

  @*/
void GtableOutputFree( Gtable_output *output )
{
    if (output->free_fcn) 
	(output->free_fcn)( output );
    else
	free( output );
}
