#include <stdio.h>
#include "gtable.h"
#include "gtableimpl.h"
#include <stdlib.h>
#include <string.h>

/*
 * Simple XML output.  Each column has a separate format string.
 */

typedef struct {
    FILE *file;
    const char *newline;
    int  cols;                    /* Number of defined column strings */
    char **col_fmt;               /* Array of column formats */
    char *start_row, *end_row;    /* Start and end row strings */
    char *start_tbl, *end_tbl;    /* Start and end table strings */
} gxml_data_t;

static int gi_nop( Gtable *ctx )
{
    return 0;
}

static int gi_start_table( Gtable *ctx, Gtable_output *output, 
			  void *extra_data )
{
    gxml_data_t *gt = (gxml_data_t *)(output->extra_data);
    if (!gt->start_tbl) return 1;
    fputs( gt->start_tbl, gt->file );
    fputs( gt->newline, gt->file );
    return 0;
}

static int gi_end_table( Gtable *ctx, void *extra_data )
{
    gxml_data_t *gt = (gxml_data_t *)extra_data;
    if (!gt->end_tbl) return 1;
    fputs( gt->end_tbl, gt->file );
    fputs( gt->newline, gt->file );
    return 0;
}

static int gi_start_row( Gtable *ctx, Gtable_entry *row, int row_num, 
		      void *extra_data )
{
    gxml_data_t *gt = (gxml_data_t *)extra_data;
    if (!gt->start_row) return 1;
    fputs( gt->start_row, gt->file );
    return 0;
}
static int gi_end_row( Gtable *ctx, Gtable_entry *row, int row_num, 
		      void *extra_data )
{
    gxml_data_t *gt = (gxml_data_t *)extra_data;
    if (!gt->end_row) return 1;
    fputs( gt->end_row, gt->file );
    fputs( gt->newline, gt->file );
    return 0;
}

/* Compute the alignment strings just like HTML version in case we want to use 
   them later 
*/
static int gi_output_col( Gtable *ctx, Gtable_entry *col, 
			 int row_num, int col_num, void *extra_data )
{
    gxml_data_t *gt = (gxml_data_t *)extra_data;
    char        *fmt;

#ifdef XML_USE_ALIGNMENT
    const char *align_string, *valign_string;

    switch (col->halign) {
    case GTABLE_LEFT:
	align_string = "LEFT"; 	 break;
    case GTABLE_RIGHT:
	align_string = "RIGHT";	 break;
    case GTABLE_CENTER:
    case GTABLE_DEFAULT:
	align_string = "CENTER"; break;
    }

    switch (col->valign) {
    case GTABLE_TOP:
	valign_string = "TOP"; break;
    case GTABLE_BOTTOM:
	valign_string = "BOTTOM"; break;
    case GTABLE_VDEFAULT:
    case GTABLE_VCENTER:
	valign_string = "CENTER"; break;
    }
#endif

    if (col_num <= gt->cols) 
	fmt = gt->col_fmt[col_num];
    else
	fmt = "%s";

    fprintf( gt->file, fmt, col->contents );
    return 0;
}

static void gi_free( Gtable_output *output )
{
    gxml_data_t *gt = (gxml_data_t *)output->extra_data;
    int i;

    for (i=0; i<gt->cols; i++) 
	free( gt->col_fmt[i] );
    free( gt->col_fmt );
    free( output->extra_data );
    free( output );
}

/* 
   Nulls for file and newline give the defaults (stdout and \n).
   setting newline to "\r\n" makes the output easier to read on Windows 
   systems.
*/
Gtable_output *Gtable_CreateXML( FILE *file, const char *newline )
{
    Gtable_output *new;
    gxml_data_t  *gt;

    new = (Gtable_output *)malloc( sizeof(Gtable_output) );
    if (!new) return NULL;

    new->start_table     = gi_start_table;
    new->end_table       = gi_end_table;

    new->start_row       = gi_start_row;
    new->end_row         = gi_end_row;
    new->output_col      = gi_output_col;
    new->output_multicol = (Gtable_col_fcn)gi_nop;
    new->free_fcn        = gi_free;

    gt                   = (gxml_data_t *)malloc( sizeof(gxml_data_t) );
    if (!gt) return 0;
    gt->file             = file ? file : stdout;
    gt->newline          = newline ? strdup(newline) : strdup( "\n" );
    gt->cols             = 0;
    gt->col_fmt          = 0;
    gt->start_row        = 0;
    gt->end_row          = 0;
    gt->start_tbl        = 0;
    gt->end_tbl          = 0;
    new->extra_data      = gt;

    return new;
}


/*
 * Set the XML strings
 */
int Gtable_XML_Names( Gtable_output *output, const char *start_table,
		      const char *end_table, const char *start_row,
		      const char *end_row )
{
    gxml_data_t *gt = (gxml_data_t *)output->extra_data;

    gt->start_tbl = strdup( start_table );
    gt->end_tbl   = strdup( end_table );
    gt->start_row = strdup( start_row );
    gt->end_row   = strdup( end_row );

    if (gt->start_tbl && gt->end_tbl && gt->start_row && gt->end_row)
	return 0;
    return 1;
}

int Gtable_XML_Column_names( Gtable *gtable, Gtable_output *output, int col, 
			     const char *col_fmt )
{
    gxml_data_t *gt = (gxml_data_t *)output->extra_data;
    int i;

    if (col + 1> gt->cols) {
	if (gt->cols == 0) {
	    if (gtable && gtable->cols > col) gt->cols = gtable->cols;
	    else                              gt->cols = col + 1;
	    gt->col_fmt = ( char ** )malloc( gt->cols * sizeof(char * ) );
	    if (!gt->col_fmt) return 1;
	    for (i=0; i<gt->cols; i++) 
		gt->col_fmt[i] = 0;
	}
	else {
	    /* Copy the old list to a new list */
	    char ** new_ptr;
	    /* Could try realloc, but we'd need the rest of this code 
	       anyway... */
	    new_ptr = (char **)malloc( (col + 1)* sizeof(char *) );
	    if (!new_ptr) return 1;
	    for (i=0; i<gt->cols; i++)
		new_ptr[i] = gt->col_fmt[i];
	    for (i=gt->cols; i<=col; i++) 
		new_ptr[i] = 0;
	    free( gt->col_fmt );
	    gt->col_fmt = new_ptr;
	    gt->cols = col + 1;
	}
    }
    if (gt->col_fmt[col]) {
	free( gt->col_fmt[col] );
    }
    gt->col_fmt[col] = strdup( col_fmt );
    return 0;
}
	
