/* -*- Mode: C; c-basic-offset:4 ; -*- */
/*
 *
 *  (C) 2001-2003 by Argonne National Laboratory.
 *      See COPYRIGHT in top-level directory.
 */
/* 
   Changes 11/17/2003:
      + buschelm - Added logging of calls to MPI_Comm_dup and MPI_Comm_split
        with additional logging when comm used is MPI_IDENT to 
        MPI_COMM_WORLD and MPI_COMM_SELF

   Changes 10/27/2003:
      + buschelm - Improved output for synchronizing routines.

   Changes 9/32/2003:
      + WDG - Introduced msgsize_t to allow the use of 64 bit integers
        for message lengths when available.

   Changes 9/22/2003:
      + WDG - Added MPI_Pcontrol hook.  Fixed some collective routines
        so that only data *sent* is recorded (more uniform, since 
        amount sent and amount received is different.  Sent defined
	as "provided", e.g., bcast sends the data once from the root.
	Added explanatory header, and added more controls over output
	format.  Added synctime control.

   Changes 9/17/2003:
      + WDG - Many changes.  Fixed broken table output.  Restored
        ability to output in multiple output formats (not just XML),
        though only text currently available
	Changed to structures for the data collection to reduce
	overhead and simplify code.
	Split a lot of long code into separate routines; moved global
	declarations to the group of routines that use them.  We may
	want to subdivide the file at some point.

   Changes 9/15/2003:
      + WDG - Many updates, including new configure and fpmpiconf.h file

   Changes 10/22/2001:
      + Added nearest-neighbor collection info. Each proc will
        record the number of times it communicates with others as
        well as keep a running sum of the size of messages it
        sends to others.

        This is rather kludgey at the moment.

   Changes 7/20/2001:
      + Added support for proposed XML schema

      "It is assumed that each performance record must contain the 
      following (consistent with the second approach for performance 
      events [1]):
       - Record type (identifies the record)
       - Measurement container (consists of one or more measurements)
       - Metadata (system and application attributes that provide the 
         context in which the performance data was collected)
       - Timestamp (identify when the record was created)

   Changes 6/15/2001:
      + Added gtable output

   Changes 5/3/2000:
      + Send sizes were converted to doubles due to overflows.

   Changes 4/13/2000:
      + Output format was changed to reflect needs of the Terascale
        initiative.

   Changes 3/13/2000:
      + We now print the min/max/avg min_loc/max_loc for the communication
        partners across all ranks, instead of printing the list of all 
	ranks which participated in a point-to-point communication.
        In a 5-pt stencil, for instance, the avg. should be close to 5
	with min/max values differing due to boundary effects.
	The heading for this section (in MPI_PROFILE_FILE) has changed
	to "Communication Partners".

   Changes 3/3/2000:
      + On Linux systems, the /proc/(pid)/stat file is read in order
        to get memory usage info. Lines are notated with a 
	#ifdef LINUX directive.
      + We now check for the T5_EVENT1 environment variable, since
        if this is set, it generates an error (conflicting hwc error)
	on SGI machines when using the hardware counters. If set, an error
	message is printed to STDERR and executions halts.
	If not set, we check for the presence of the MPI_PROFILE_EVENT1
	environment variable. If set to 21, floating point ops are polled 
	from the hardware counters, and if set to 26,  L2 cache misses 
	are polled. Other values are allowed since we don't check this
	value. If MPI_PROFILE_EVENT1 is not set, L2 cache misses are
	polled by default.
      + The destination/sources of each point-to-point communication
        call is tracked and printed out, under the heading
	of "Processor Communication Data". 
	We just keep track of the ranks, independent of the communicators.  
	We *should* use MPI_Group_translate_ranks to make all ranks 
	relative to MPI_COMM_WORLD; in another realease...

      o Still to do: The Fortran interfaces for the MI_Waitxxx routines
	are incorrect and need to be fixed.

   Changes 2/24/2000:
      + Totals for point-to-point sends:
          total calls, size of messages, time. This includes
	  the usual min/max/avg min-loc/max-loc info as well.
	  Also, the communication rate (size/time) is calculated.
      + Total for collective sends: (same as above)
      + Total "barrier" time is shown for collective routines
        (this is the time spent waiting for nodes to initiate 
        collective sends/receives)

   Changes 2/22/2000:
      + added profiling for MPI_Waitxxx calls (C only)
      + stats printed for Waits & Barrier calls.
      + added more MPI routines (Sendrecv_replace, Pack/Unpack)
*/

#include "fpmpiconf.h"
#include "profiler.h"
#include "mpi.h"

#ifdef GTABLE
#include "gtable.h"
#endif /* GTABLE */

#include <fcntl.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef HAVE_SYS_RESOURCE_H
#include <sys/resource.h>
#endif
#include <time.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
/* Global defs & variables */

#define ONE_MB 1048576
#define ONE_KB 1024
#define MAX_MSG 1073741824
#define MPI_PROF_FILE_NAME "fpmpi_profile"

#define MPI_PROF_FLOPS 21
#define MPI_PROF_L2 26


/* timing control */
static int profControl      = 1;
static int profControlSync  = 1;
static int profControlDests = 1;

/* timing */
static clock_t MPI_Start_time, MPI_End_time;

static double startwtime, endwtime; /* These get overwritten as needed */
static double sync_time;            /* also used as needed */

/* Message size bins */
static int bin_id;

enum bin_sizes {sz_0, sz_1to4, sz_5to8, sz_9to16, sz_17to32, sz_33to64, 
		sz_65to128, sz_129to256, sz_257to512, sz_513to1024, 
		sz_1to4K, sz_5to8K, sz_9to16K, sz_17to32K, sz_33to64K, 
		sz_65to128K, sz_129to256K, sz_257to512K, sz_513to1024K,
		sz_1to4M, sz_5to8M, sz_9to16M, sz_17to32M, sz_33to64M,
		sz_65to128M, sz_129to256M, sz_257to512M, sz_513to1024M, 
		sz_large};
/*
static char *bin_names[NBINS] =  {"0 ", "1 to 4 ", "5 to 8 ", "9 to 16 ", 
			   "17 to 32 ", "33 to 64 ", "65 to 128 ", 
			   "129 to 256 ", "257 to 512 ", "513 to 1024 ", 
			   "1 to 4 K", "5 to 8 K", "9 to 16 K",
			   "17 to 32 K", "33 to 64 K", "65 to 128 K", 
			   "129 to 256 K", "257 to 512 K", "513 to 1024 K",
			   "1 to 4 M", "5 to 8 M", "9 to 16 M", "17 to 32 M", 
			   "33 to 64 M", "65 to 128 M", "129 to 256 M", 
			   "257 to 512 M", "513 to 1024 M", "> 1024 M"};
*/
/* Usage of this is commented out in putil.c */
/* char *max_bin_names[NBINS] =   */
/*       {"0B ", "0-4B ", "4-8B ", "8-16B ", "16-32B ", */
/*        "32-64B ", "64-128B ", "128-256B ", "256-512B ",  */
/*        "512-1023B ", "1024B-4KB ", "4-8KB ", "8-16KB ", */
/*        "16-32KB ", "32-64KB ", "64-128KB ", */
/*        "128-256KB ", "256-512KB ", "512-1024KB ", */
/*        "1024KB-4MB ", "4-8MB ", "8-16MB ", */
/*        "16-32MB ", "32-64MB ", "64-128MB", */
/*        "128-256MB ", "256-512MB ", "512-1024MB ",  */
/*        ">1024MB "}; */

/* 
 * Definitions for the data that is collected 
 */
typedef struct {
    msgsize_t size, size2;    /* sum of exact size and square of size */
    double time, time2;       /* sum of time and square of time */
    int    calls;             /* Number of calls at this size */
} CommBinData; 
/* FIXME: Some of the code relies on the partner data array (?) */
#ifndef GET_DETAILED_PARTNER_DATA
#define GET_DETAILED_PARTNER_DATA
#endif
#ifdef GET_DETAILED_PARTNER_DATA
typedef struct {
    double times;
    long   sizes;
    int    partners;
} CommPartnerData;
#endif
typedef struct {
    CommBinData     b[NBINS];     /* One bin for each range of message size */
#ifdef GET_DETAILED_PARTNER_DATA
    CommPartnerData p[MAX_PROCS]; /* One for each potential partner */
#endif
} CommData;

#define COUNTER_DECL(name) static CommData name##_data

#define COUNTER_DECL_COLL(name) \
    COUNTER_DECL(name); \
    static double name##_timesync[NBINS]

COUNTER_DECL_COLL(allgather);
COUNTER_DECL_COLL(allgatherv);
COUNTER_DECL_COLL(allreduce);
COUNTER_DECL_COLL(alltoall);
COUNTER_DECL_COLL(alltoallv);

/* Barrier, wait, test, Comm_dup, and Comm_split are special */

static int    barrier_calls = 0;

static int    wait_calls = 0;
static int    waitall_calls = 0;
static int    waitany_calls = 0;
static int    waitsome_calls = 0;

static int    test_calls = 0;
static int    testall_calls = 0;
static int    testany_calls = 0;
static int    testsome_calls = 0;

static double barrier_time = 0;

static double wait_time = 0;
static double waitall_time = 0;
static double waitany_time = 0;
static double waitsome_time = 0;

static double test_time = 0;
static double testall_time = 0;
static double testany_time = 0;
static double testsome_time = 0;

static int    comm_dup_calls        = 0;
static int    comm_world_dups       = 0;
static int    comm_self_dups        = 0;
static int    comm_dup_world_keyval = MPI_KEYVAL_INVALID;
static int    comm_dup_self_keyval  = MPI_KEYVAL_INVALID;
static int    comm_dup_other_keyval = MPI_KEYVAL_INVALID;

static int    comm_split_calls  = 0;
static int    comm_world_splits = 0;

COUNTER_DECL(bcast);
COUNTER_DECL(gather);
COUNTER_DECL(gatherv);
COUNTER_DECL(reduce_scatter);
COUNTER_DECL(reduce);
COUNTER_DECL(scan);
COUNTER_DECL(scatter);
COUNTER_DECL(scatterv);

COUNTER_DECL(recv);
COUNTER_DECL(send);
COUNTER_DECL(sendrecv);
COUNTER_DECL(sendrecv_replace);
COUNTER_DECL(bsend);
COUNTER_DECL(ibsend);
COUNTER_DECL(isend);
COUNTER_DECL(issend);
COUNTER_DECL(irsend);
COUNTER_DECL(rsend);
COUNTER_DECL(ssend);
COUNTER_DECL(irecv);

COUNTER_DECL(pack);
COUNTER_DECL(unpack);

static int    iprobe_calls = 0;
static double iprobe_time  = 0.;
static int    probe_calls  = 0;
static double probe_time   = 0.;

static int    total_isend_calls = 0;
static msgsize_t total_isend_size = 0;
static double total_isend_time = 0;
/* experimental */
static msgsize_t isend_avg_size = 0;
static long int total_isend_comm_sizes[MAX_PROCS];
static double total_isend_comm_times[MAX_PROCS];
static long int total_isend_comm_partners[MAX_PROCS];

static int    total_bsend_calls = 0;
static msgsize_t total_bsend_size = 0;
static double total_bsend_time = 0;
/* experimental */
static msgsize_t bsend_avg_size = 0;
static long int total_bsend_comm_sizes[MAX_PROCS];
static double total_bsend_comm_times[MAX_PROCS];
static long int total_bsend_comm_partners[MAX_PROCS];

static int    total_col_send_calls = 0;
static msgsize_t total_col_send_size = 0;
static double total_col_send_time = 0;
/* experimental */
static msgsize_t col_send_avg_size = 0;
static long int total_col_send_comm_sizes[MAX_PROCS];
static double total_col_send_comm_times[MAX_PROCS];
static long int total_col_send_comm_partners[MAX_PROCS];

#if defined(ENABLE_STDDEV)
/* experimental */
static msgsize_t isend_stddev_size = 0;
static msgsize_t bsend_stddev_size = 0;
static msgsize_t col_send_stddev_size = 0;
#endif

/* Bit vectors to keep track of which processes this one has been in
   communication with. */
/* On Linux (gcc) ints are 32-bit */
/* Include a way to keep track of the amount of data sent to each
   destination.  
   FIXME: an alternative is to hash on the destination, trading a 
   small amount of additional time with much smaller data area */
#define COMM_DESTS(name) \
    static int32_t name##_target[MAX_PROCBY32];\
    static msgsize_t name##_targetsize[MAX_PROCS][NBINS];

COMM_DESTS(recv);
COMM_DESTS(send);
COMM_DESTS(sendrecv);
COMM_DESTS(sendrecv_replace);
COMM_DESTS(bsend);
COMM_DESTS(ibsend);
COMM_DESTS(isend);
COMM_DESTS(issend);
COMM_DESTS(irsend);
COMM_DESTS(rsend);
COMM_DESTS(ssend);
COMM_DESTS(irecv);

/* Experimental: keep track of which procs we communicate with */
/* This should also use comm_translate_ranks for comm != MPI_COMM_WORLD,
   and use a small (probably single element) cache for the group of
   another communicator.  The comm cache is also freed whenever MPI_Comm_free
   is called. 
   FIXME:
   This should also be divided up into a high-resolution but space-hungry
   version and a low-resolution but space-efficient version.  The 
   space-efficient version just keeps a bit for each partner.  The
   high-resolution version keeps track of communication to each partner.
*/
static MPI_Comm comm_cache = MPI_COMM_NULL;
static MPI_Group group_cache = MPI_GROUP_NULL;
static MPI_Group group_world = MPI_GROUP_NULL;

#ifdef GET_DETAILED_PARTNER_DATA
#define UPDATE_COMM_DESTS(name,comm,source) if (profControlDests){ \
    int _s = source; \
    if (comm != MPI_COMM_WORLD) {\
        if (comm != comm_cache) {\
            if (group_cache != MPI_GROUP_NULL) PMPI_Group_free(&group_cache);\
            PMPI_Comm_group( comm, &group_cache );\
            comm_cache = comm;\
            }\
        PMPI_Group_translate_ranks( group_cache, 1, &source, group_world, &_s);\
     }\
    {\
    CommPartnerData *const _p = &(name##_data.p[_s]);\
    _p->partners ++; _p->sizes += buf_size; \
    _p->times += (endwtime - startwtime);\
     name##_target[_s >> 5] |= (1 << (_s & 0x01F));\
     name##_targetsize[_s][bin_id]+=buf_size;}}
#else
#define UPDATE_COMM_DESTS(name,comm,source) if (profControlDests){ \
    int _s = source; \
    if (comm != MPI_COMM_WORLD) {\
        if (comm != comm_cache) {\
            if (group_cache != MPI_GROUP_NULL) PMPI_Group_free(&group_cache);\
            PMPI_Comm_group( comm, &group_cache );\
            comm_cache = comm;\
            }\
        PMPI_Group_translate_ranks( group_cache, 1, &source, group_world, &_s);\
     }\
    {\
     name##_target[_s >> 5] |= (1 << (_s & 0x01F));\
     name##_targetsize[_s][bin_id]+=buf_size;}}
#endif

#define Bit_Is_Set(target,SelectBit) \
( (target[i] & SelectBit) > 0 )

/* forward declarations */
static int Get_Bin_ID( int );
static int Get_Msg_size( int, MPI_Datatype );


/* Begin Profiling Routines */

/* This macro lets us change the way we remember the call */
#define UPDATE_COUNTER(name) if(profControl){ CommBinData *const _b = &(name##_data.b[bin_id]);\
const double _t = (endwtime - startwtime); \
 _b->time += _t;  _b->time2 += _t * _t; \
 _b->calls ++; _b->size += buf_size; _b->size2 += buf_size * buf_size;}

#define UPDATE_COUNTER_SYNC(name) \
    name##_timesync[bin_id] += sync_time;

/* Break out the time sync but leave the separate update macro
o   in case we decide to change how we handle the data recording */
#define UPDATE_COUNTER_COLL(name) UPDATE_COUNTER(name)

/* Convert a (count/datatype) into a size in bytes */
static int Get_Msg_size( int count, MPI_Datatype datatype )
{
    int dsize;
    
    PMPI_Type_size( datatype, &dsize );
    return count * dsize;
}

/* ------------------------------------------------------------------------
 * Service routines for collecting data
 * ------------------------------------------------------------------------
 */
/* Determine message bin ID.
   To keep this fast, the bin id's are a power of two; we can compute the
   bin id with a quick tree code.  STILL TO DO!  FIXME

   We should really just look for 4 * 2^k (since most scientific data
   will be at least a multiple of the size of an int).

   size >>= 2;   // divide by four 
   Find the location of the largest bit.
   See http://graphics.stanford.edu/~seander/bithacks.html#IntegerLog
   for an example.  An easy code is

   int b = buf_size;
   while (b >>=1) bin_id ++;
   if (bin_id > 1) bin_id -= 1;  // combine 1 and 2,3 byte messages
   
   This gives:
   bin_id == 0  Messages of 0 bytes
   bin_id == 1  Messages of 1,2,3 bytes
   bin_id == 2  Messages of 4-7 bytes
   bin_id == k  Messages of 2^k - 2^k-1 bytes
*/
static int Get_Bin_ID(int buf_size)
{
  int bin_id = 0; /* By default, assume a zero size message */

  /* We subdivide this a little to catch the short messages 
     early */
  if (buf_size > 4*ONE_KB) {
      if ( buf_size > MAX_MSG ) bin_id = NBINS - 1; /* The last bin is for
						       messages > 1024MB */
      else if ( buf_size > 512*ONE_MB ) bin_id = sz_513to1024M;
      else if ( buf_size > 256*ONE_MB ) bin_id = sz_257to512M;
      else if ( buf_size > 128*ONE_MB ) bin_id = sz_129to256M;
      else if ( buf_size >  64*ONE_MB ) bin_id = sz_65to128M;
      else if ( buf_size >  32*ONE_MB ) bin_id = sz_33to64M;
      else if ( buf_size >  16*ONE_MB ) bin_id = sz_17to32M;
      else if ( buf_size >   8*ONE_MB ) bin_id = sz_9to16M;
      else if ( buf_size >   4*ONE_MB ) bin_id = sz_5to8M;
      else if ( buf_size >     ONE_MB ) bin_id = sz_1to4M;
      else if ( buf_size > 512*ONE_KB ) bin_id = sz_513to1024K;
      else if ( buf_size > 256*ONE_KB ) bin_id = sz_257to512K;
      else if ( buf_size > 128*ONE_KB ) bin_id = sz_129to256K;
      else if ( buf_size >  64*ONE_KB ) bin_id = sz_65to128K;
      else if ( buf_size >  32*ONE_KB ) bin_id = sz_33to64K;
      else if ( buf_size >  16*ONE_KB ) bin_id = sz_17to32K;
      else if ( buf_size >   8*ONE_KB ) bin_id = sz_9to16K;
      else /*if ( buf_size >   4*ONE_KB )*/ bin_id = sz_5to8K;
  }
  else {
      if      ( buf_size >     ONE_KB ) bin_id = sz_1to4K;
      else if ( buf_size > 512 )        bin_id = sz_513to1024;
      else if ( buf_size > 256 )        bin_id = sz_257to512;
      else if ( buf_size > 128 )        bin_id = sz_129to256;
      else if ( buf_size >  64 )        bin_id = sz_65to128;
      else if ( buf_size >  32 )        bin_id = sz_33to64;
      else if ( buf_size >  16 )        bin_id = sz_17to32;
      else if ( buf_size >   8 )        bin_id = sz_9to16;
      else if ( buf_size >   4 )        bin_id = sz_5to8;
      else if ( buf_size >   0  )       bin_id = sz_1to4;
  }
  return bin_id;
}


/*
 * --------------------------------------------------------------------
 * Here begins the replacements for the MPI routines
 * --------------------------------------------------------------------
 */
/*
    MPI_Allgather - prototyping replacement for MPI_Allgather
    Trace the beginning and ending of MPI_Allgather.
*/
int MPI_Allgather( void *sendbuf, int sendcount, MPI_Datatype sendtype, 
		   void *recvbuf, int recvcount, MPI_Datatype recvtype, 
		   MPI_Comm comm )
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( sendcount, sendtype );
  bin_id = Get_Bin_ID( buf_size );

  if (profControlSync) {
      startwtime = MPI_Wtime();
      PMPI_Barrier( comm );
      sync_time = MPI_Wtime() - startwtime;
      UPDATE_COUNTER_SYNC(allgather);
  }

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Allgather( sendbuf, sendcount, sendtype, recvbuf, 
			      recvcount, recvtype, comm );
  endwtime = MPI_Wtime();
  UPDATE_COUNTER_COLL(allgather);

  return returnVal;
}

/*
    MPI_Allgatherv - prototyping replacement for MPI_Allgatherv
    Trace the beginning and ending of MPI_Allgatherv.
*/
int MPI_Allgatherv( void *sendbuf, int sendcount, MPI_Datatype sendtype, 
		    void *recvbuf, int *recvcounts, int *displs, 
		    MPI_Datatype recvtype, MPI_Comm comm ) 
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( sendcount, sendtype );
  bin_id = Get_Bin_ID( buf_size);

  if (profControlSync) {
      startwtime = MPI_Wtime();
      PMPI_Barrier( comm );
      sync_time = MPI_Wtime() - startwtime;
      UPDATE_COUNTER_SYNC(allgatherv);
  }

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Allgatherv( sendbuf, sendcount, sendtype, recvbuf, 
			       recvcounts, displs, recvtype, comm );
  endwtime = MPI_Wtime();
  UPDATE_COUNTER_COLL(allgatherv);
  
  return returnVal;
}

/*
    MPI_Allreduce - prototyping replacement for MPI_Allreduce
    Trace the beginning and ending of MPI_Allreduce.
*/
int MPI_Allreduce( void *sendbuf, void *recvbuf, int count, 
		   MPI_Datatype datatype, MPI_Op op, MPI_Comm comm )
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( count, datatype );
  bin_id = Get_Bin_ID( buf_size );

  if (profControlSync) {
      startwtime = MPI_Wtime();
      PMPI_Barrier( comm );
      sync_time = MPI_Wtime() - startwtime;
      UPDATE_COUNTER_SYNC(allreduce);
  }

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Allreduce( sendbuf, recvbuf, count, datatype, op, comm );
  endwtime = MPI_Wtime();
  UPDATE_COUNTER_COLL(allreduce);

  return returnVal;
}

/*
    MPI_Alltoall - prototyping replacement for MPI_Alltoall
    Trace the beginning and ending of MPI_Alltoall.
*/
int MPI_Alltoall( void *sendbuf, int sendcount, MPI_Datatype sendtype, 
		  void *recvbuf, int recvcnt, MPI_Datatype recvtype, 
		  MPI_Comm comm )
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( sendcount, sendtype );
  bin_id = Get_Bin_ID( buf_size );

  if (profControlSync) {
      startwtime = MPI_Wtime();
      PMPI_Barrier( comm );
      sync_time = MPI_Wtime() - startwtime;
      UPDATE_COUNTER_SYNC(alltoall);
  }

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Alltoall( sendbuf, sendcount, sendtype, recvbuf,
			     recvcnt, recvtype, comm );
  endwtime = MPI_Wtime();
  UPDATE_COUNTER_COLL(alltoall);
  
  return returnVal;
}

/*
    MPI_Alltoallv - prototyping replacement for MPI_Alltoallv
    Trace the beginning and ending of MPI_Alltoallv.
*/
int MPI_Alltoallv( void *sendbuf, int *sendcnts, int *sdispls, 
		   MPI_Datatype sendtype, void *recvbuf, int *recvcnts, 
		   int *rdispls, MPI_Datatype recvtype, MPI_Comm comm )
{
  int returnVal;
  int buf_size, i, nprocs;

  buf_size = 0;
  PMPI_Comm_size( comm, &nprocs );
  for (i=0; i<nprocs; i++) {
      buf_size += Get_Msg_size( sendcnts[i], sendtype );
  }
  bin_id = Get_Bin_ID(buf_size);  /* Determine message bin ID */

  if (profControlSync) {
      startwtime = MPI_Wtime();
      PMPI_Barrier( comm );
      sync_time = MPI_Wtime() - startwtime;
      UPDATE_COUNTER_SYNC(alltoallv);
  }

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Alltoallv( sendbuf, sendcnts, sdispls, sendtype, recvbuf,
			      recvcnts, rdispls, recvtype, comm );
  endwtime = MPI_Wtime();
  UPDATE_COUNTER_COLL(alltoallv);
  
  return returnVal;
}

/*
    MPI_Barrier - prototyping replacement for MPI_Barrier
    Trace the beginning and ending of MPI_Barrier.
*/
int MPI_Barrier( MPI_Comm comm )
{
  int returnVal;

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Barrier( comm );
  endwtime = MPI_Wtime();
  barrier_time  += endwtime - startwtime;
  barrier_calls += 1;

  return returnVal;
}

/*
     MPI_Iprobe - prototyping replacement for MPI_Iprobe
     Count calls, and accumulate total time.
*/
int MPI_Iprobe(int source, int tag, MPI_Comm comm, int *flag, MPI_Status *status)
{
  int returnVal;
 
  startwtime    = MPI_Wtime();
  returnVal     = PMPI_Iprobe(source,tag,comm,flag,status);
  endwtime      = MPI_Wtime();
  iprobe_time  += endwtime - startwtime;
  iprobe_calls += 1;

  return returnVal;
}

/*
     MPI_Probe - prototyping replacement for MPI_Probe
     Count calls, and accumulate total time.
*/
int MPI_Probe(int source, int tag, MPI_Comm comm, MPI_Status *status)
{
  int returnVal;
 
  startwtime    = MPI_Wtime();
  returnVal     = PMPI_Probe(source,tag,comm,status);
  endwtime      = MPI_Wtime();
  probe_time  += endwtime - startwtime;
  probe_calls += 1;

  return returnVal;
}

/*
    MPI_Bcast - prototyping replacement for MPI_Bcast
    Trace the beginning and ending of MPI_Bcast.
*/
int MPI_Bcast( void *buffer, int count, MPI_Datatype datatype, 
	       int root, MPI_Comm comm )
{
  int returnVal;
  int buf_size, rank;

  PMPI_Comm_rank( comm, &rank );
  buf_size = Get_Msg_size( count, datatype );/* Size is in bytes */
  bin_id = Get_Bin_ID(buf_size);  /* Determine message bin ID */
  if (rank != 0) {
      /* This sets the bin_id to the size of message being received
	 but counts as no data *sent.* */
      buf_size = 0;
  }

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Bcast( buffer, count, datatype, root, comm );
  endwtime = MPI_Wtime();
  UPDATE_COUNTER(bcast);

  return returnVal;
}

/*
    MPI_Comm_dup - duplicates of MPI_COMM_WORLD and MPI_COMM_SELF are counted
*/
int MPI_Comm_dup( MPI_Comm oldcomm, MPI_Comm *newcomm )
{
  int returnVal;
  int *dummyattrval, flag;

  if (comm_dup_world_keyval == MPI_KEYVAL_INVALID) {
      MPI_Keyval_create(MPI_DUP_FN,MPI_NULL_DELETE_FN,&comm_dup_world_keyval,(void *)0);
  }
  if (comm_dup_self_keyval  == MPI_KEYVAL_INVALID) {
      MPI_Keyval_create(MPI_DUP_FN,MPI_NULL_DELETE_FN,&comm_dup_self_keyval,(void *)0);
  }
  if (comm_dup_other_keyval == MPI_KEYVAL_INVALID) {
      MPI_Keyval_create(MPI_DUP_FN,MPI_NULL_DELETE_FN,&comm_dup_other_keyval,(void *)0);
  }

  MPI_Attr_get(oldcomm,comm_dup_other_keyval,(void**)&dummyattrval,&flag);
  if (!flag) {
      MPI_Attr_get(oldcomm,comm_dup_world_keyval,(void**)&dummyattrval,&flag);
      if (flag) {
          comm_world_dups += 1;
      } else {
          MPI_Attr_get(oldcomm,comm_dup_self_keyval,(void**)&dummyattrval,&flag);
          if (flag) {
              comm_self_dups += 1;
          } else {
              PMPI_Comm_size(oldcomm,&flag);
              if (flag == 1) {
                  MPI_Attr_put(oldcomm,comm_dup_self_keyval,(void *)&comm_dup_self_keyval);
                  comm_self_dups += 1;
              } else {
                  PMPI_Comm_compare(oldcomm,MPI_COMM_WORLD,&flag);
                  if (flag == MPI_IDENT) {
                      MPI_Attr_put(oldcomm,comm_dup_world_keyval,(void *)&comm_dup_world_keyval);
                      comm_world_dups += 1;
                  } else {
                      MPI_Attr_put(oldcomm,comm_dup_other_keyval,(void *)&comm_dup_other_keyval);
                  }
              }
          } 
      }
  }

  returnVal = PMPI_Comm_dup(oldcomm,newcomm);

  comm_dup_calls += 1;
  return returnVal;
}

/*
    MPI_Comm_split - splits of MPI_COMM_WORLD and MPI_COMM_SELF are counted
*/
int MPI_Comm_split( MPI_Comm oldcomm, int color, int key, MPI_Comm *newcomm )
{
  int returnVal;
  int *dummyattrval, flag;
  if (comm_dup_world_keyval == MPI_KEYVAL_INVALID) {
      MPI_Keyval_create(MPI_DUP_FN,MPI_NULL_DELETE_FN,&comm_dup_world_keyval,(void *)0);
  }
  if (comm_dup_other_keyval == MPI_KEYVAL_INVALID) {
      MPI_Keyval_create(MPI_DUP_FN,MPI_NULL_DELETE_FN,&comm_dup_other_keyval,(void *)0);
  }

  MPI_Attr_get(oldcomm,comm_dup_other_keyval,(void**)&dummyattrval,&flag);
  if (!flag) {
      MPI_Attr_get(oldcomm,comm_dup_world_keyval,(void**)&dummyattrval,&flag);
      if (flag) {
          comm_world_splits += 1;
      } else {
          PMPI_Comm_compare(oldcomm,MPI_COMM_WORLD,&flag);
          if (flag == MPI_IDENT) {
              MPI_Attr_put(oldcomm,comm_dup_world_keyval,(void *)&comm_dup_world_keyval);
              comm_world_splits += 1;
          } else {
              MPI_Attr_put(oldcomm,comm_dup_other_keyval,(void *)&comm_dup_other_keyval);
          }
      }
  }
  returnVal = PMPI_Comm_split(oldcomm,color,key,newcomm);
  comm_split_calls += 1;
  return returnVal;
}

/*
    MPI_Gather - prototyping replacement for MPI_Gather
    Trace the beginning and ending of MPI_Gather.
*/
int MPI_Gather( void *sendbuf, int sendcnt, MPI_Datatype sendtype, 
		void *recvbuf, int recvcount, MPI_Datatype recvtype, 
		int root, MPI_Comm comm )
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( sendcnt, sendtype );
  bin_id = Get_Bin_ID(buf_size);
  
  startwtime = MPI_Wtime();
  returnVal = PMPI_Gather( sendbuf, sendcnt, sendtype, recvbuf, 
			   recvcount, recvtype, root, comm );
  endwtime = MPI_Wtime();
  UPDATE_COUNTER(gather);

  return returnVal;
}

/*
    MPI_Gatherv - prototyping replacement for MPI_Gatherv
    Trace the beginning and ending of MPI_Gatherv.
*/
int MPI_Gatherv( void *sendbuf, int sendcnt, MPI_Datatype sendtype, 
		 void *recvbuf, int *recvcnts, int *displs, 
		 MPI_Datatype recvtype, int root, MPI_Comm comm )
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( sendcnt, sendtype );
  bin_id = Get_Bin_ID(buf_size);
  
  startwtime = MPI_Wtime();
  returnVal = PMPI_Gatherv( sendbuf, sendcnt, sendtype, recvbuf, recvcnts,
			    displs, recvtype, root, comm );
  endwtime = MPI_Wtime();
  UPDATE_COUNTER(gatherv);
  
  return returnVal;
}

/*
    MPI_Pack - prototyping replacement for MPI_Pack
    Log the beginning and ending of the time spent in MPI_Pack calls.
*/
int MPI_Pack( void *inbuf, int incount, MPI_Datatype datatype, void *outbuf, 
	      int outcount, int *position, MPI_Comm comm )
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( incount, datatype );
  bin_id = Get_Bin_ID( buf_size );
  
  startwtime = MPI_Wtime();
  returnVal  = PMPI_Pack(inbuf, incount, datatype, outbuf, outcount, 
			position, comm );
  endwtime   = MPI_Wtime();
  UPDATE_COUNTER(pack);

  return returnVal;
}

/*
    MPI_Unpack - prototyping replacement for MPI_Unpack
    Log the beginning and ending of the time spent in MPI_Unpack calls.
*/
int MPI_Unpack( void *inbuf, int insize, int *position, void *outbuf, 
		int outcount, MPI_Datatype datatype, MPI_Comm comm )
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( outcount, datatype );
  bin_id = Get_Bin_ID( buf_size );
  
  startwtime = MPI_Wtime();
  returnVal = PMPI_Unpack( inbuf, insize, position, outbuf, outcount, 
			   datatype, comm );
  endwtime = MPI_Wtime();
  UPDATE_COUNTER(unpack);

  return returnVal;
}

/*
    MPI_Reduce_scatter - prototyping replacement for MPI_Reduce_scatter
    Trace the beginning and ending of MPI_Reduce_scatter.
*/
int MPI_Reduce_scatter( void *sendbuf, void *recvbuf, int *recvcnts, 
			MPI_Datatype datatype, MPI_Op op, MPI_Comm comm )
{
  int returnVal;
  int ierr, buf_size;
  
  buf_size = Get_Msg_size( recvcnts[0], datatype ); /* FIXME */
  bin_id = Get_Bin_ID(buf_size);

  if (profControlSync) {
      startwtime = MPI_Wtime();	  
      ierr	 = PMPI_Barrier( comm );
      endwtime	 = MPI_Wtime();	  
      sync_time  = startwtime - endwtime;
  }

  startwtime = endwtime;
  returnVal = PMPI_Reduce_scatter( sendbuf, recvbuf, recvcnts, datatype,
				   op, comm );
  endwtime = MPI_Wtime();	
  UPDATE_COUNTER(reduce_scatter);  
  
  return returnVal;
}

/*
    MPI_Reduce - prototyping replacement for MPI_Reduce
    Trace the beginning and ending of MPI_Reduce.
*/
int MPI_Reduce( void *sendbuf, void *recvbuf, int count, 
		  MPI_Datatype datatype, MPI_Op op, int root, 
		  MPI_Comm comm )
{
  int returnVal;
  int buf_size;
  
  buf_size = Get_Msg_size( count, datatype );
  bin_id = Get_Bin_ID(buf_size);

  startwtime = endwtime;
  returnVal = PMPI_Reduce( sendbuf, recvbuf, count, datatype, op, root, comm );
  endwtime = MPI_Wtime();	
  UPDATE_COUNTER(reduce);  

  return returnVal;
}

/*
    MPI_Scan - prototyping replacement for MPI_Scan
    Trace the beginning and ending of MPI_Scan.
*/
int MPI_Scan( void *sendbuf, void *recvbuf, int count, 
		MPI_Datatype datatype, MPI_Op op, MPI_Comm comm )
{
  int returnVal;
  int buf_size;
  
  buf_size = Get_Msg_size( count, datatype );
  bin_id = Get_Bin_ID(buf_size);

  startwtime = endwtime;
  returnVal = PMPI_Scan( sendbuf, recvbuf, count, datatype, op, comm );
  endwtime = MPI_Wtime();
  UPDATE_COUNTER(scan);	  

  return returnVal;
}

/*
    MPI_Scatter - prototyping replacement for MPI_Scatter
    Trace the beginning and ending of MPI_Scatter.
*/
int MPI_Scatter( void *sendbuf, int sendcnt, MPI_Datatype sendtype, 
		 void *recvbuf, int recvcnt, MPI_Datatype recvtype, 
		 int root, MPI_Comm comm )
{
  int returnVal;
  int buf_size, rank;

  buf_size = Get_Msg_size( sendcnt, sendtype );
  bin_id = Get_Bin_ID(buf_size);
  PMPI_Comm_rank( comm, &rank );
  if (rank != root) buf_size = 0;  /* See MPI_Bcast */

  startwtime = endwtime;
  returnVal = PMPI_Scatter( sendbuf, sendcnt, sendtype, recvbuf, recvcnt, 
			    recvtype, root, comm );
  endwtime = MPI_Wtime();	  
  UPDATE_COUNTER(scatter);

  return returnVal;
}

/*
    MPI_Scatterv - prototyping replacement for MPI_Scatterv
    Trace the beginning and ending of MPI_Scatterv.
*/
int MPI_Scatterv( void *sendbuf, int *sendcnts, int *displs, 
		  MPI_Datatype sendtype, void *recvbuf, int recvcnt, 
		  MPI_Datatype recvtype, int root, MPI_Comm comm )
{
  int returnVal;
  int buf_size, i, nprocs, rank;

  PMPI_Comm_rank( comm, &rank );
  PMPI_Comm_size( comm, &nprocs );
  if (rank == root) {
      buf_size = 0;
      for (i=0; i<nprocs; i++) {
	  buf_size += Get_Msg_size( sendcnts[i], sendtype );
      }
      bin_id = Get_Bin_ID(buf_size);
  }
  else {
      /* Compute the bin based on the receive size */
      buf_size = Get_Msg_size( recvcnt, recvtype );
      bin_id = Get_Bin_ID(buf_size);
      buf_size = 0;   /* see MPI_Bcast */
  }

  startwtime = endwtime;
  returnVal = PMPI_Scatterv( sendbuf, sendcnts, displs, sendtype, recvbuf, 
			     recvcnt, recvtype, root, comm );
  endwtime = MPI_Wtime();	
  UPDATE_COUNTER(scatterv);  

  return returnVal;
}

/*
    MPI_Recv - prototyping replacement for MPI_Recv
    Trace the beginning and ending of MPI_Recv.
*/
int MPI_Recv( void *buf, int count, MPI_Datatype datatype, int source, 
	       int tag, MPI_Comm comm, MPI_Status *status )
{
  int returnVal;
  int buf_size;
  int actsrc;
  MPI_Status tmpstatus;
  MPI_Status *sp = status;

  if (status == MPI_STATUS_IGNORE) sp = &tmpstatus;
  
  startwtime = MPI_Wtime();
  returnVal = PMPI_Recv( buf, count, datatype, source, tag, comm, sp );
  endwtime = MPI_Wtime();

  MPI_Get_count( sp, datatype, &count );
  actsrc = sp->MPI_SOURCE;
  buf_size = Get_Msg_size( count, datatype );
  bin_id = Get_Bin_ID(buf_size);  /* Determine message bin ID */
  
  UPDATE_COUNTER(recv);
  UPDATE_COMM_DESTS(recv, comm, actsrc);

  return returnVal;
}


/*
    MPI_Send - prototyping replacement for MPI_Send
    Trace the beginning and ending of MPI_Send.
*/
int MPI_Send( void *buf, int count, MPI_Datatype datatype, int dest, 
	       int tag, MPI_Comm comm )
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( count, datatype );
  bin_id = Get_Bin_ID(buf_size);

  startwtime = MPI_Wtime();	  
  returnVal  = PMPI_Send( buf, count, datatype, dest, tag, comm );
  endwtime   = MPI_Wtime();	  
  UPDATE_COUNTER(send);
  UPDATE_COMM_DESTS(send, comm, dest);

  return returnVal;
}

/*
    MPI_Sendrecv - prototyping replacement for MPI_Sendrecv
    Trace the beginning and ending of MPI_Sendrecv.
*/
int MPI_Sendrecv( void * sendbuf, int sendcount, MPI_Datatype sendtype, 
		   int dest, int sendtag, void * recvbuf, int recvcount, 
		   MPI_Datatype recvtype, int source, int recvtag, 
		   MPI_Comm comm, MPI_Status * status)
{
  int returnVal;
  int buf_size;
  /* int actsrc; */
  MPI_Status tmpstatus;
  MPI_Status *sp = status;

  if (status == MPI_STATUS_IGNORE) sp = &tmpstatus;

  buf_size = Get_Msg_size( sendcount, sendtype );
  bin_id = Get_Bin_ID(buf_size);

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Sendrecv( sendbuf, sendcount, sendtype, dest, sendtag, 
			     recvbuf, recvcount, recvtype, source, recvtag,
			     comm, sp );
  endwtime   = MPI_Wtime();	  
  UPDATE_COUNTER(sendrecv);
  UPDATE_COMM_DESTS(sendrecv, comm, dest); /* Only has send dest */

  return returnVal;
}

/*
    MPI_Sendrecv_replace - prototyping replacement for MPI_Sendrecv_replace
    Log the beginning and ending of the time spent in MPI_Sendrecv_replace calls.
*/
int MPI_Sendrecv_replace( void *buf, int count, MPI_Datatype datatype, 
			  int dest, int sendtag, int source, int recvtag, 
			  MPI_Comm comm, MPI_Status *status )
{
  int returnVal;
  int buf_size;
  MPI_Status tmpstatus;
  MPI_Status *sp = status;

  if (status == MPI_STATUS_IGNORE) sp = &tmpstatus;

  buf_size = Get_Msg_size( count, datatype );
  bin_id = Get_Bin_ID(buf_size);

  startwtime = MPI_Wtime();	  
  returnVal  = PMPI_Sendrecv_replace( buf, count, datatype, dest, 
			 sendtag, source, recvtag, comm, sp );
  endwtime   = MPI_Wtime();	  
  UPDATE_COUNTER(sendrecv_replace);
  UPDATE_COMM_DESTS(sendrecv_replace, comm, dest);
  
  return returnVal;
}

/*
    MPI_Bsend - prototyping replacement for MPI_Bsend
    Trace the beginning and ending of MPI_Bsend.
*/
int MPI_Bsend( void *buf, int count, MPI_Datatype datatype, int dest, 
	       int tag, MPI_Comm comm )
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( count, datatype );
  bin_id = Get_Bin_ID(buf_size);

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Bsend( buf, count, datatype, dest, tag, comm );
  endwtime   = MPI_Wtime();	  
  UPDATE_COUNTER(bsend);
  UPDATE_COMM_DESTS(bsend, comm, dest);

  return returnVal;
}

/*
    MPI_Ibsend - prototyping replacement for MPI_Ibsend
    Trace the beginning and ending of MPI_Ibsend.
*/
int MPI_Ibsend( void *buf, int count, MPI_Datatype datatype, int dest, 
		int tag, MPI_Comm comm, MPI_Request *request )
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( count, datatype );
  bin_id = Get_Bin_ID(buf_size);
  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Ibsend( buf, count, datatype, dest, tag, comm, request );
  endwtime   = MPI_Wtime();	  
  UPDATE_COUNTER(ibsend);
  UPDATE_COMM_DESTS(ibsend, comm, dest);

  return returnVal;
}

/*
    MPI_Irecv - prototyping replacement for MPI_Irecv
    Trace the beginning and ending of MPI_Irecv.
*/
int MPI_Irecv( void * buf, int count, MPI_Datatype datatype, int source,
		int tag, MPI_Comm comm, MPI_Request * request)
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( count, datatype );
  bin_id = Get_Bin_ID(buf_size);
  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Irecv( buf, count, datatype, source, tag, comm, request );
  endwtime   = MPI_Wtime();	  
  UPDATE_COUNTER(irecv);
  if (source >= 0) {
      /* FIXME: Must ignore ANY_SOURCE until Wait/Test */
      UPDATE_COMM_DESTS(irecv, comm, source);
  }

  return returnVal;
}

/*
    MPI_Irsend - prototyping replacement for MPI_Irsend
    Trace the beginning and ending of MPI_Irsend.
*/
int MPI_Irsend( void *buf, int count, MPI_Datatype datatype, int dest, 
		int tag, MPI_Comm comm, MPI_Request *request )
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( count, datatype );
  bin_id = Get_Bin_ID(buf_size);

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Irsend( buf, count, datatype, dest, tag, comm, request );
  endwtime   = MPI_Wtime();	  
  UPDATE_COUNTER(irsend);
  UPDATE_COMM_DESTS(irsend, comm, dest);

  return returnVal;
}

/*
    MPI_Rsend - prototyping replacement for MPI_Rsend
    Trace the beginning and ending of MPI_Rsend.
*/
int MPI_Rsend( void * buf, int count, MPI_Datatype datatype, int dest,
	       int tag, MPI_Comm comm)
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( count, datatype );
  bin_id = Get_Bin_ID(buf_size);

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Rsend( buf, count, datatype, dest, tag, comm );
  endwtime   = MPI_Wtime();	
  UPDATE_COUNTER(rsend);  
  UPDATE_COMM_DESTS(rsend, comm, dest);

  return returnVal;
}

/*
    MPI_Ssend - prototyping replacement for MPI_Ssend
    Trace the beginning and ending of MPI_Ssend.
*/
int MPI_Ssend( void * buf, int count, MPI_Datatype datatype, int dest,
		int tag, MPI_Comm comm)
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( count, datatype );
  bin_id = Get_Bin_ID(buf_size);

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Ssend( buf, count, datatype, dest, tag, comm );
  endwtime   = MPI_Wtime();	
  UPDATE_COUNTER(ssend);  
  UPDATE_COMM_DESTS(ssend, comm, dest);

  return returnVal;
}
/*
    MPI_Isend - prototyping replacement for MPI_Isend
    Trace the beginning and ending of MPI_Isend.
*/
int MPI_Isend( void *buf, int count, MPI_Datatype datatype, int dest, 
		int tag, MPI_Comm comm, MPI_Request *request )
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( count, datatype );
  bin_id = Get_Bin_ID(buf_size);

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Isend( buf, count, datatype, dest, tag, comm, request );
  endwtime   = MPI_Wtime();	  
  UPDATE_COUNTER(isend);
  UPDATE_COMM_DESTS(isend, comm, dest);

  return returnVal;
}

/*
    MPI_Issend - prototyping replacement for MPI_Issend
    Trace the beginning and ending of MPI_Issend.
*/
int MPI_Issend( void * buf, int count, MPI_Datatype datatype, 
		 int dest, int tag, MPI_Comm comm, MPI_Request * request)
{
  int returnVal;
  int buf_size;

  buf_size = Get_Msg_size( count, datatype );
  bin_id = Get_Bin_ID(buf_size);

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Issend( buf, count, datatype, dest, tag, comm, request );
  endwtime   = MPI_Wtime();	  
  UPDATE_COUNTER(issend);
  UPDATE_COMM_DESTS(issend, comm, dest);

  return returnVal;
}

/*
    MPI_Wait - prototyping replacement for MPI_Wait
    Log the beginning and ending of the time spent in MPI_Wait calls.
*/
int MPI_Wait( MPI_Request *request, MPI_Status *status )
{
  int returnVal;

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Wait( request, status );
  endwtime = MPI_Wtime();
  wait_time  += endwtime - startwtime;
  wait_calls += 1;

  return returnVal;
}

/*
    MPI_Waitall - prototyping replacement for MPI_Waitall
    Log the beginning and ending of the time spent in MPI_Waitall calls.
*/
int MPI_Waitall( int count, MPI_Request *array_of_requests, 
		 MPI_Status *array_of_statuses )
{
  int returnVal;
  
  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Waitall( count, array_of_requests, array_of_statuses );
  endwtime = MPI_Wtime();
  waitall_time  += endwtime - startwtime;
  waitall_calls += 1;

  return returnVal;
}

/*
    MPI_Waitany - prototyping replacement for MPI_Waitany
    Log the beginning and ending of the time spent in MPI_Waitany calls.
*/
int MPI_Waitany( int count, MPI_Request *array_of_requests, 
		 int *index, MPI_Status *status )
{
  int returnVal;

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Waitany( count, array_of_requests, index, status );
  endwtime = MPI_Wtime();
  waitany_time  += endwtime - startwtime;
  waitany_calls += 1;

  return returnVal;
}

/*
    MPI_Waitsome - prototyping replacement for MPI_Waitsome
    Log the beginning and ending of the time spent in MPI_Waitsome calls.
*/
int MPI_Waitsome( int incount, MPI_Request *array_of_requests, 
		  int *outcount, int *array_of_indices, 
		  MPI_Status *array_of_statuses )
{
  int returnVal;

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Waitsome( incount, array_of_requests, outcount, 
			     array_of_indices, array_of_statuses );
  endwtime = MPI_Wtime();
  waitsome_time  += endwtime - startwtime;
  waitsome_calls += 1;

  return returnVal;
}

int   MPI_Test( MPI_Request *request, int *flag, MPI_Status *status )
{
  int   returnVal;

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Test( request, flag, status );
  endwtime = MPI_Wtime();
  test_time  += endwtime - startwtime;
  test_calls += 1;

  return returnVal;
}

int  MPI_Testall( int count, MPI_Request *array_of_requests, int *flag, 
		  MPI_Status *array_of_statuses )
{
  int  returnVal;

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Testall( count, array_of_requests, flag, array_of_statuses );
  endwtime = MPI_Wtime();
  testall_time  += endwtime - startwtime;
  testall_calls += 1;

  return returnVal;
}

int  MPI_Testany( int count, MPI_Request *array_of_requests, int *index, 
		  int *flag, MPI_Status *status )
{
  int  returnVal;

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Testany( count, array_of_requests, index, flag, status );
  endwtime = MPI_Wtime();
  testany_time  += endwtime - startwtime;
  testany_calls += 1;

  return returnVal;
}

int  MPI_Testsome( int incount, MPI_Request *array_of_requests, 
		   int *outcount, int *array_of_indices, 
		   MPI_Status *array_of_statuses )
{
  int  returnVal;

  startwtime = MPI_Wtime();	  
  returnVal = PMPI_Testsome( incount, array_of_requests, outcount, 
			     array_of_indices, array_of_statuses );
  endwtime = MPI_Wtime();
  testsome_time  += endwtime - startwtime;
  testsome_calls += 1;

  return returnVal;
}


/*
 * --------------------------------------------------------------------
 * Here begins the definition of routines that are intercepted to 
 * support other operations (such as the translations to ranks in comm world)
 * --------------------------------------------------------------------
 */
int MPI_Comm_free( MPI_Comm *comm )
{
    /* flush the cache of known communicators */
    comm_cache = MPI_COMM_NULL;
    return PMPI_Comm_free( comm );
}
/* ------------------------------------------------------------------------
 * Here is the definition of MPI_Pcontrol
 *
 * MPI_Pcontrol( flag ) has the following definition
 *  flag has three bits.  The lower (0x1) bit controls the collection
 *  of data. The second (0x2) bit controls the collection of 
 *  time synchronization data for the collective routines
 *  The third (0x4) bit controls the collection of data about destinations.
 * ------------------------------------------------------------------------
 */
int MPI_Pcontrol( const int flag, ... )
{
    profControl      = (flag & 0x1);
    profControlSync  = (flag & 0x2) == 0x2;
    profControlDests = (flag & 0x4) == 0x4;
    return MPI_SUCCESS;
}

/*
 * --------------------------------------------------------------------
 * Here begins the defintiion of INIT
 * --------------------------------------------------------------------
 */
void fpmpiInitPerformanceCounters(void);
/* Program name */
static char *cmdname = 0;

/* Control the output comments */
static int Verbose = 0;

/* What sort of output will we generate */
typedef enum { OUT_TEXT, OUT_XML, OUT_LATEX } OutputForm;

static OutputForm outform = OUT_TEXT;

/* How much output will we generate? */
static int FullOutput = 1;

/*
    MPI_Init - prototyping replacement for MPI_Init
    Trace the beginning and ending of MPI_Init.
*/
int MPI_Init( int *argc, char ***argv )
{
  int returnVal;
  int llrank;

  MPI_Start_time = clock();
  
  returnVal = PMPI_Init( argc, argv );

  if (argv && *argv) cmdname = (*argv)[0];

  PMPI_Comm_rank( MPI_COMM_WORLD, &llrank );

  /* Get the group of comm world so that we can translate ranks in other
     communicators */
  PMPI_Comm_group( MPI_COMM_WORLD, &group_world );

  fpmpiInitPerformanceCounters();

  /* Get the output form */
  { char *p = getenv( "FPMPI_OUT_FORM" );
      outform = OUT_TEXT;
      if (p) {
	  if (strcmp( p, "XML" ) == 0) outform = OUT_XML;
	  else if (strcmp( p, "LATEX" ) == 0) outform = OUT_LATEX;
      }
  }

  return returnVal;
}

/*
 * --------------------------------------------------------------------
 * Here begins the defintiion of FINALIZE
 * --------------------------------------------------------------------
 */
/* Forward references */
/* FIXME: Make these static, or move them to another file and make them
   fpmpiOutputxxx */
FILE *OutputOpenFile( int );
void OutputProgramDescription( FILE *, int, int );
void OutputResourceUsage( FILE *, int, int  );
void OutputCommunicationData( FILE *, int, int );
void OutputPartners( FILE *, int );
void OutputHeader( FILE *, int );
void OutputTail( FILE *, int );
void Output_TestWait( FILE *, int, int );
void Output_DupSplit( FILE *, int);
/*
    MPI_Finalize - prototyping replacement for MPI_Finalize
    Trace the beginning and ending of MPI_Finalize.
    
    Here is the order of events:

    DetermineOutputType (e.g., text, xml)
    GetOutputFilename
    OutputProgramDesc (e.g., cmdname, date, overall times)
    OutputResourceUsage
    OutputCommunicationData 
        This uses the "generic table" (gtable) routines
    OutputPartners

*/
int MPI_Finalize(  )
{
  int returnVal;
  int llrank, nprocs;
  FILE *pf;

  PMPI_Comm_rank( MPI_COMM_WORLD, &llrank );
  PMPI_Comm_size( MPI_COMM_WORLD, &nprocs );

  pf = OutputOpenFile( llrank );

  OutputHeader( pf, llrank );

  OutputProgramDescription( pf, llrank, nprocs );

  OutputResourceUsage( pf, llrank, nprocs );

  (void)PMPI_Barrier( MPI_COMM_WORLD );

  OutputCommunicationData( pf, llrank, nprocs );
  Output_TestWait( pf, llrank, nprocs );
  Output_DupSplit( pf, llrank );

  OutputPartners( pf, llrank );

  OutputTail( pf, llrank );
  /* OutputTail closes the file */
  
  returnVal = PMPI_Finalize();
  return returnVal;
}

/*
 * --------------------------------------------------------------------
 * Here begins the output routines
 * --------------------------------------------------------------------
 */
static void printout_msgstats_coll( FILE *, 
				    CommData *, double *, const char *,
				    int, int );
static void printout_comm_dests( int32_t *, const char *, void * );
static void printout_total_msgstats( FILE *, int, int);
#ifdef UNUSED
static void printout_header( FILE *, const char *, int );
#endif

static void getrange_double( const double[], int, 
			     double [], double [], double [], int [], int [] );

static void getrange_int_single(int, int *, int *,int *, int *, int *t);
static void getrange_double_single(double, double *, int *, double *, 
				   int *, double *);
#if SIZEOF_LONG_LONG >= 8
static void getrange_msgsize_single(msgsize_t, msgsize_t *, 
				    int *, msgsize_t *, 
				    int *, msgsize_t *);
#else
#define getrange_msgsize_single getrange_double_single
#endif

/* Printout the data for one communication item, based on the 
   data in cd and timesync 
   
   Given the data, compute the max, min, and average over all processes.
   Add to the current Gtable

*/
void CBCombineMax( void *, void *, int *, MPI_Datatype * );
void CBCombineMin( void *, void *, int *, MPI_Datatype * );
void CBCombineSum( void *, void *, int *, MPI_Datatype * );

void CBCombineMax( void *cptrin, void *cptrout, 
		   int *count, MPI_Datatype *dtype )
{
  int i, n = *count;
  CommBinData *cin = (CommBinData *)cptrin;
  CommBinData *cout = (CommBinData *)cptrout;
  for (i=0; i<n; i++) {
    if (cout[i].size < cin[i].size)
      cout[i].size = cin[i].size;
    if (cout[i].size2 < cin[i].size2)
      cout[i].size2 = cin[i].size2;
    if (cout[i].time < cin[i].time)
      cout[i].time = cin[i].time;
    if (cout[i].time2 < cin[i].time2)
      cout[i].time2 = cin[i].time2;
    if (cout[i].calls < cin[i].calls)
      cout[i].calls = cin[i].calls;
  }
}
/* For the min, we may want to ignore min values that have no
   associated calls */
/* Question: do we want to know if ANYONE had zero calls? */
void CBCombineMin( void *cptrin, void *cptrout,
		   int *count, MPI_Datatype *dtype )
{
    int i, n = *count;
    CommBinData *cin = (CommBinData *)cptrin;
    CommBinData *cout = (CommBinData *)cptrout;
    for (i=0; i<n; i++) {
	if (cout[i].calls == 0) {
	    /* just copy input*/
	    cout[i].calls = cin[i].calls;
	    cout[i].size  = cin[i].size;
	    cout[i].size2 = cin[i].size2;
	    cout[i].time  = cin[i].time;
	    cout[i].time2 = cin[i].time2;
	}
	else {
	    /* Only update if cin had some calls */
	    if (cin[i].calls > 0) {
		if (cout[i].size > cin[i].size)
		    cout[i].size = cin[i].size;
		if (cout[i].size2 > cin[i].size2)
		    cout[i].size2 = cin[i].size2;
		if (cout[i].time > cin[i].time)
		    cout[i].time = cin[i].time;
		if (cout[i].time2 > cin[i].time2)
		    cout[i].time2 = cin[i].time2;
		if (cout[i].calls > cin[i].calls)
		    cout[i].calls = cin[i].calls;
	    }
	}
    }
}

void CBCombineSum( void *cptrin, void *cptrout, 
		   int *count, MPI_Datatype *dtype )
{
  int i, n = *count;
  CommBinData *cin = (CommBinData *)cptrin;
  CommBinData *cout = (CommBinData *)cptrout;
  for (i=0; i<n; i++) {
    cout[i].size += cin[i].size;
    cout[i].size2 += cin[i].size2;
    cout[i].time += cin[i].time;
    cout[i].time2 += cin[i].time2;
    cout[i].calls += cin[i].calls;
  }
}

#define MAX_TIME 100000
#if SIZEOF_LONG_LONG >= 8
#define MAX_LEN  (((long long)1)<<40)
#else
#define MAX_LEN  (1<<30)
#endif

void printout_msgstats_coll( FILE *pf, 
			     CommData *cd, double *timesync, const char *name,
			     int llrank, int nprocs )
{
  static MPI_Op cb_max, cb_min, cb_sum = MPI_OP_NULL;
  static MPI_Datatype cb_dtype = MPI_DATATYPE_NULL;
  CommBinData  cbMax[NBINS], cbMin[NBINS], cbSum[NBINS];
  char   percentSent[NBINS+1];
  double timesyncmax[NBINS], timesyncmin[NBINS], timesyncsum[NBINS];

  if (cb_sum == MPI_OP_NULL) {
    PMPI_Op_create( CBCombineMax, 1, &cb_max );
    PMPI_Op_create( CBCombineMin, 1, &cb_min );
    PMPI_Op_create( CBCombineSum, 1, &cb_sum );
  }
  if (cb_dtype == MPI_DATATYPE_NULL) {
    MPI_Datatype din[4];
    MPI_Aint displs[4];
    int blk[4], i;

#if SIZEOF_LONG_LONG >= 8
    PMPI_Address( &cbMax[0].size, &displs[0] );
    PMPI_Address( &cbMax[0].time, &displs[1] );
    PMPI_Address( &cbMax[0].calls, &displs[2] );
    PMPI_Address( &cbMax[1].size, &displs[3] );
    for (i=3; i>=0; i--) 
      displs[i] -= displs[0];
    blk[0] = 2; din[0] = MPI_LONG_LONG;
    blk[1] = 2; din[1] = MPI_DOUBLE;
    blk[2] = 1; din[2] = MPI_INT;
    blk[3] = 1; din[3] = MPI_UB;
    PMPI_Type_struct( 4, blk, displs, din, &cb_dtype );
#else    
    PMPI_Address( &cbMax[0].size, &displs[0] );
    PMPI_Address( &cbMax[0].calls, &displs[1] );
    PMPI_Address( &cbMax[1].size, &displs[2] );
    for (i=2; i>=0; i--) 
      displs[i] -= displs[0];
    blk[0] = 4; din[0] = MPI_DOUBLE;
    blk[1] = 1; din[1] = MPI_INT;
    blk[2] = 1; din[2] = MPI_UB;
    PMPI_Type_struct( 3, blk, displs, din, &cb_dtype );
#endif
    PMPI_Type_commit( &cb_dtype );
  }

  PMPI_Reduce( cd->b, cbSum, NBINS, cb_dtype, cb_sum, 0, MPI_COMM_WORLD );
  PMPI_Reduce( cd->b, cbMax, NBINS, cb_dtype, cb_max, 0, MPI_COMM_WORLD );
  PMPI_Reduce( cd->b, cbMin, NBINS, cb_dtype, cb_min, 0, MPI_COMM_WORLD );

  /* All processes will have timesync if any do */
  if (timesync) {
      /* We could reduce these to a single reduce with a user-defined
	 operation if it becomes important.  Since few routines
	 include this measurement, this simpler code is better */
      PMPI_Reduce( timesync, timesyncmax, NBINS, MPI_DOUBLE, MPI_MAX, 0, 
		   MPI_COMM_WORLD );
      PMPI_Reduce( timesync, timesyncmin, NBINS, MPI_DOUBLE, MPI_MIN, 0,
		   MPI_COMM_WORLD );
      PMPI_Reduce( timesync, timesyncsum, NBINS, MPI_DOUBLE, MPI_SUM, 0,
		   MPI_COMM_WORLD );
  }

  if (llrank == 0) {
      double avg_time, max_time, min_time;
      msgsize_t avg_size, max_size, min_size;
      int    total_calls, i;
      msgsize_t total_size;
    /* Compute the average message size, time, and total sent */
    avg_size    = 0;
    avg_time    = 0;
    total_calls = 0;
    total_size  = 0;
    min_time    = MAX_TIME;
    max_time    = 0;
    min_size    = MAX_LEN;
    max_size    = 0;
    for (i=0; i<NBINS; i++) {
	/* FIXME: if there were no calls at a particular size, do NOT 
	   take the min as zero */
      total_calls += cbSum[i].calls;
      total_size  += cbSum[i].size;
      avg_time    += cbSum[i].time;
      if (min_size > cbMin[i].size)
	min_size = cbMin[i].size;
      if (max_size < cbMax[i].size)
	max_size = cbMax[i].size;
      if (min_time > cbMin[i].time) 
	min_time = cbMin[i].time;
      if (max_time < cbMax[i].time)
	max_time = cbMax[i].time;
    }

    avg_time /= nprocs;
    avg_size = total_size / nprocs;

    if (total_calls > 0) {
	/* name calls avg/min/max time avg/min/max len */
	/* Older version printed
	   name mincalls, maxcalls, avcalls, process with min, 
	   process with max,
	   min time, maxtime, av time, process with min , process with max 
	*/
#if SIZEOF_LONG_LONG >= 8 
      fprintf( pf, "%32s\t%d\t%f\t%f\t%f\t%lld\t%lld\t%lld", 
	       name, total_calls, avg_time, min_time, max_time,
	       (long long)avg_size, (long long)min_size, (long long)max_size );
#else
      fprintf( pf, "%32s\t%d\t%f\t%f\t%f\t%f\t%f\t%f", 
	       name, total_calls, avg_time, min_time, max_time,
	       avg_size, min_size, max_size );
#endif

      if (timesync) {
	  double totalsync, minsync, maxsync;
	  /* Print out the synchronization time */
	  totalsync = 0; 
	  minsync   = MAX_TIME;
	  maxsync   = 0;
	  for (i=0; i<NBINS; i++) {
	      totalsync += timesyncsum[i];
	      if (cbSum[i].calls > 0 && timesyncmin[i] < minsync) 
		  minsync = timesyncmin[i];
	      if (timesyncmax[i] > maxsync) 
		  maxsync = timesyncmax[i];
	  }
	  fprintf( pf, "\t%f\t%f\t%f", 
		   totalsync/nprocs, minsync, maxsync );
      }
      /* Printout the distribution by bin.  Compute the fraction
	 sent (by calls) in units of 10%, by calls. */
      for (i=0; i<NBINS; i++) {
	  /* The total_calls/2-1 is for rounding */
	  int npercent = (10 * cbSum[i].calls + total_calls/2-1)/ total_calls;
	  /* The >= test is just a check on something bad */
	  if (npercent < 0) {
	      /* Check for total_calls < 3 */
	      if (total_calls < 3) {
		  npercent = (10 * cbSum[i].calls) / total_calls;
	      }
	  }
	  if (npercent < 10 && npercent > 0)
	      percentSent[i] = '0' + npercent;
	  else if (npercent == 0) {
	      /* Distinguish between no calls and few calls */
	      if (cbSum[i].calls)
		  percentSent[i] = '.';
	      else
		  percentSent[i] = '0';
	  }
	  else {
	      percentSent[i] = '*'; 
	  }
      }
      percentSent[i] = 0;
      fprintf( pf, "\t%s", percentSent );

      /* Eventually, the above should use the gtable code to 
	 ensure that the output can be output in any form */

      /* Printout the distribution by bin.  Compute the fraction
	 sent (by total bytes) in units of 10%, by calls. */
      for (i=0; i<NBINS; i++) {
	  /* The total_size/2-1 is for rounding */
	  int npercent = (10 * cbSum[i].size + total_size/2-1)/ total_size;
	  /* The >= test is just a check on something bad */
	  if (npercent < 0) {
	      /* Check for total_size < 3 */
	      if (total_size < 3) {
		  npercent = (10 * cbSum[i].size) / total_size;
	      }
	  }
	  if (npercent < 10 && npercent > 0)
	      percentSent[i] = '0' + npercent;
	  else if (npercent == 0) {
	      /* Distinguish between no data and just a little bit of data */
	      if (cbSum[i].size)
		  percentSent[i] = '.';
	      else
		  percentSent[i] = '0';
	  }
	  else {
	      percentSent[i] = '*'; 
	  }
      }
      percentSent[i] = 0;
      fprintf( pf, "\t%s", percentSent );

      /* Eventually, the above should use the gtable code to 
	 ensure that the output can be output in any form */

      /* Write the end of line */
      fprintf( pf, "\n" );

    }

    /* FIXME: Also gather partner information; output 
       as the number of calls to each partner */
  }
}

/* Output all of the test/wait/barrier information */
static void printout_testwait( FILE *pf, int llrank, int nprocs, int calls, 
			       double time, const char *name )
{
    int maxcalls, mincalls, total_calls, maxloc, minloc;
    double max_time, min_time, sum_time, avg_time;

    getrange_int_single( calls, &maxcalls, &maxloc, &mincalls, &minloc,
			 &total_calls );
    getrange_double_single( time, &max_time, &maxloc, 
			    &min_time, &minloc, &sum_time );
    
    if (llrank == 0) {
	if (calls > 0) {
	    avg_time = sum_time / nprocs;
	    fprintf( pf, "%32s\t%d\t%f\t%f\t%f\n",
	       name, total_calls, avg_time, min_time, max_time );
	}
    }
}
/*barrier, wait, waitall, waitany, waitsome, test, testall, testany, testsome */
void Output_TestWait( FILE *pf, int llrank, int nprocs )
{

    printout_testwait( pf, llrank, nprocs, barrier_calls,  barrier_time,  "MPI_Barrier" );
    printout_testwait( pf, llrank, nprocs, wait_calls,     wait_time,     "MPI_Wait" );
    printout_testwait( pf, llrank, nprocs, waitall_calls,  waitall_time,  "MPI_Waitall" );
    printout_testwait( pf, llrank, nprocs, waitany_calls,  waitany_time,  "MPI_Waitany" );
    printout_testwait( pf, llrank, nprocs, waitsome_calls, waitsome_time, "MPI_Waitsome" );
    printout_testwait( pf, llrank, nprocs, test_calls,     test_time,     "MPI_Test" );
    printout_testwait( pf, llrank, nprocs, testall_calls,  testall_time,  "MPI_Testall" );
    printout_testwait( pf, llrank, nprocs, testany_calls,  testany_time,  "MPI_Testany" );
    printout_testwait( pf, llrank, nprocs, testsome_calls, testsome_time, "MPI_Testsome" );
    printout_testwait( pf, llrank, nprocs, iprobe_calls,   iprobe_time,   "MPI_Iprobe" );
    printout_testwait( pf, llrank, nprocs, probe_calls,    probe_time,    "MPI_Probe" );

}

static void printout_dupsplit( FILE *pf, int llrank, int calls, const char *name)
{
    int maxcalls, mincalls, total_calls, maxloc, minloc;

    getrange_int_single(calls, &maxcalls, &maxloc, &mincalls, &minloc, &total_calls);
    if (llrank == 0) {
        if (calls > 0) {
            fprintf( pf, "%32s\t%d\n", name, total_calls);
        }
    }
}
void Output_DupSplit( FILE *pf, int llrank)
{
    printout_dupsplit(pf, llrank, comm_dup_calls,    "MPI_Comm_dup");
    printout_dupsplit(pf, llrank, comm_world_dups,   "MPI_Comm_dup(MPI_COMM_WORLD)");
    printout_dupsplit(pf, llrank, comm_self_dups,    "MPI_Comm_dup(MPI_COMM_SELF)");
    printout_dupsplit(pf, llrank, comm_split_calls,  "MPI_Comm_split");
    printout_dupsplit(pf, llrank, comm_world_splits, "MPI_Comm_split(MPI_COMM_WORLD)");
}


/* 
   The following collective routines compute the 
   min/max/rank of min/rank of max 
   of an array of values, returning the values in supplied arrays.

   These are three essentially identical routines, except for the choice of 
   input datatype (int, long, or double) and the related structures and
   MPI datatypes. 
 */


static void getrange_double( const double vals[], int count, 
			   double sumout[], double minout[], double maxout[], 
			     int minloc[], int maxloc[] )
{
    typedef struct { double val; int loc; } doubleint;
    doubleint *lvals, *gvals;
    int i, nprocs, rank;

    lvals = (doubleint *)malloc( 2 * count * sizeof(doubleint) );
    gvals = lvals + count;

    PMPI_Comm_size( MPI_COMM_WORLD, &nprocs );
    PMPI_Comm_rank( MPI_COMM_WORLD, &rank );
    for (i=0; i<count; i++) {
	lvals[i].val = vals[i];
	lvals[i].loc = rank;
    }
    PMPI_Reduce( lvals, gvals, count, MPI_DOUBLE_INT, MPI_MAXLOC, 
		 0, MPI_COMM_WORLD );
    for (i=0; i<count; i++) {
	maxout[i] = gvals[i].val;
	maxloc[i] = gvals[i].loc;
    }
    PMPI_Reduce( lvals, gvals, count, MPI_DOUBLE_INT, MPI_MINLOC, 
		 0, MPI_COMM_WORLD );
    for (i=0; i<count; i++) {
	minout[i] = gvals[i].val;
	minloc[i] = gvals[i].loc;
    }
    PMPI_Reduce( (double *)vals, sumout, count, MPI_DOUBLE, MPI_SUM, 
		 0, MPI_COMM_WORLD );
    free( lvals );
}

/* Determine the min/max/sum of single integer values across procs */
static void getrange_int_single(int val, int *maxout, int *maxloc, 
			     int *minout, int *minloc, int *sumout)
{
  typedef struct { int val; int loc; } twoint;
  twoint *lvals, *gvals;
  int nprocs, rank;

  lvals = (twoint *) malloc( 2 * sizeof(twoint) );
  gvals = lvals + 1;
  
  PMPI_Comm_size( MPI_COMM_WORLD, &nprocs );
  PMPI_Comm_rank( MPI_COMM_WORLD, &rank );

  lvals->val = val;
  lvals->loc = rank;

  PMPI_Reduce( lvals, gvals, 1, MPI_2INT, MPI_MAXLOC, 
	       0, MPI_COMM_WORLD );
  *maxout = gvals->val;
  *maxloc = gvals->loc;

  PMPI_Reduce( lvals, gvals, 1, MPI_2INT, MPI_MINLOC, 
	       0, MPI_COMM_WORLD );
  *minout = gvals->val;
  *minloc = gvals->loc;

  PMPI_Reduce( &val, sumout, 1, MPI_INT, MPI_SUM, 0, MPI_COMM_WORLD );
  free( lvals );
}

static void getrange_double_single(double val, double *maxout, int *maxloc, 
			     double *minout, int *minloc, double *sumout)
{
  typedef struct { double val; int loc; } doubleint;
  doubleint lvals, gvals;
  int nprocs, rank;

  PMPI_Comm_size( MPI_COMM_WORLD, &nprocs );
  PMPI_Comm_rank( MPI_COMM_WORLD, &rank );

  lvals.val = val;
  lvals.loc = rank;

  PMPI_Reduce( &lvals, &gvals, 1, MPI_DOUBLE_INT, MPI_MAXLOC, 
	       0, MPI_COMM_WORLD );
  *maxout = gvals.val;
  *maxloc = gvals.loc;

  PMPI_Reduce( &lvals, &gvals, 1, MPI_DOUBLE_INT, MPI_MINLOC, 
	       0, MPI_COMM_WORLD );
  *minout = gvals.val;
  *minloc = gvals.loc;

  PMPI_Reduce( &val, sumout, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD );
}

#if SIZEOF_LONG_LONG >= 8
static void getrange_msgsize_single(msgsize_t val, msgsize_t *maxout, 
				    int *maxloc, msgsize_t *minout, 
				    int *minloc, msgsize_t *sumout)
{
  typedef struct { double val; int loc; } doubleint;
  doubleint lvals, gvals;
  int nprocs, rank;
  static MPI_Datatype longlongintDtype = MPI_DATATYPE_NULL;

  if (longlongintDtype == MPI_DATATYPE_NULL) {
      /* eventually, create a longlongint type and a corresponding
	 max/min/int operation.
       */
      ;
  }

  PMPI_Comm_size( MPI_COMM_WORLD, &nprocs );
  PMPI_Comm_rank( MPI_COMM_WORLD, &rank );

  lvals.val = val;
  lvals.loc = rank;

  PMPI_Reduce( &lvals, &gvals, 1, MPI_DOUBLE_INT, MPI_MAXLOC, 
	       0, MPI_COMM_WORLD );
  *maxout = gvals.val;
  *maxloc = gvals.loc;

  PMPI_Reduce( &lvals, &gvals, 1, MPI_DOUBLE_INT, MPI_MINLOC, 
	       0, MPI_COMM_WORLD );
  *minout = gvals.val;
  *minloc = gvals.loc;

  PMPI_Reduce( &val, sumout, 1, MPI_LONG_LONG, MPI_SUM, 0, MPI_COMM_WORLD );
}
#endif

/* 
   For each node, sum the total amount of data sent and the time spent 
   in communication. Report the min/max/average for this value and
   the min/max/average communication rate as well
   
   Do the same for the "barrier" time in the collective routines, except 
   here only report the total time (no data moved).
   
   Variables beginning with total_ are totals of all BINS.
   Variables beginning with sum_ are sums across all procs.
   
   Blocking communication data is reported separtely from non-blocking.
*/

static void printout_total_msgstats(FILE *pf, int rank, int nprocs)
{
  int i;
  int  max_loc, min_loc;

  int  max_total_isend_calls, min_total_isend_calls, sum_total_isend_calls;
  msgsize_t max_total_isend_size, min_total_isend_size, sum_total_isend_size;
  double max_total_isend_time, min_total_isend_time, sum_total_isend_time;
  double comm_isend_rate, max_comm_isend_rate, min_comm_isend_rate;
  double sum_comm_isend_rate;

  int  max_total_bsend_calls, min_total_bsend_calls, sum_total_bsend_calls;
  msgsize_t max_total_bsend_size, min_total_bsend_size, sum_total_bsend_size;
  double max_total_bsend_time, min_total_bsend_time, sum_total_bsend_time;
  double comm_bsend_rate, max_comm_bsend_rate, min_comm_bsend_rate;
  double sum_comm_bsend_rate;

  int  max_total_col_send_calls, min_total_col_send_calls;
  int  sum_total_col_send_calls;
  msgsize_t max_total_col_send_size, min_total_col_send_size;
  msgsize_t sum_total_col_send_size;
  double max_total_col_send_time, min_total_col_send_time;
  double sum_total_col_send_time;
  double comm_col_rate, max_comm_col_rate, min_comm_col_rate;
  double sum_comm_col_rate;
  double total_col_timesync, min_total_col_timesync, max_total_col_timesync;
  double sum_total_col_timesync;
  msgsize_t tempstore;

  /* Initialize rates to zero just in case */
  comm_bsend_rate    = 0;
  comm_isend_rate    = 0;
  comm_col_rate      = 0;
  total_col_timesync = 0;

  for (i=0;i<NBINS;i++) {

    /* Blocking sends */
    total_bsend_calls += send_data.b[i].calls + sendrecv_data.b[i].calls 
      + sendrecv_replace_data.b[i].calls  + bsend_data.b[i].calls
      + rsend_data.b[i].calls + ssend_data.b[i].calls;
    
    total_bsend_size += send_data.b[i].size + sendrecv_data.b[i].size
      + sendrecv_replace_data.b[i].size  + bsend_data.b[i].size
      + rsend_data.b[i].size + ssend_data.b[i].size;
    
    total_bsend_time += send_data.b[i].time + sendrecv_data.b[i].time 
      + sendrecv_replace_data.b[i].time  + bsend_data.b[i].time
      + rsend_data.b[i].time + ssend_data.b[i].time;
    
    /* Immediate sends */
    total_isend_calls += ibsend_data.b[i].calls + isend_data.b[i].calls + 
	issend_data.b[i].calls
      + irsend_data.b[i].calls + bcast_data.b[i].calls + 
	gather_data.b[i].calls + gatherv_data.b[i].calls
      + reduce_data.b[i].calls + reduce_scatter_data.b[i].calls + 
	scan_data.b[i].calls
      + scatter_data.b[i].calls + scatterv_data.b[i].calls;

    total_isend_size += ibsend_data.b[i].size + isend_data.b[i].size + 
	issend_data.b[i].size
      + irsend_data.b[i].size + bcast_data.b[i].size + gather_data.b[i].size +
	gatherv_data.b[i].size
      + reduce_data.b[i].size + reduce_scatter_data.b[i].size + 
	scan_data.b[i].size
      + scatter_data.b[i].size + scatterv_data.b[i].size;

    total_isend_time += ibsend_data.b[i].time + isend_data.b[i].time + 
	issend_data.b[i].time
      + irsend_data.b[i].time + bcast_data.b[i].time + gather_data.b[i].time +
	gatherv_data.b[i].time
      + reduce_data.b[i].time + reduce_scatter_data.b[i].time + 
	scan_data.b[i].time
      + scatter_data.b[i].time + scatterv_data.b[i].time;

    /* Collective (including sync-time) */
    total_col_send_calls += allgather_data.b[i].calls + 
	allgatherv_data.b[i].calls 
      + allreduce_data.b[i].calls + alltoall_data.b[i].calls + 
	alltoallv_data.b[i].calls;
    
    total_col_send_size += allgather_data.b[i].size + 
	allgatherv_data.b[i].size 
      + allreduce_data.b[i].size + alltoall_data.b[i].size + 
	alltoallv_data.b[i].size;
    
    total_col_send_time += allgather_data.b[i].time + 
	allgatherv_data.b[i].time 
      + allreduce_data.b[i].time + alltoall_data.b[i].time + 
	alltoallv_data.b[i].time;
    
    total_col_timesync += allgather_timesync[i] + allgatherv_timesync[i]
      + allreduce_timesync[i] + alltoall_timesync[i] + alltoallv_timesync[i];

  }

  /* Experimental */
  /* Calculate average message send sizes */
  if (total_bsend_calls > 0)
      bsend_avg_size = total_bsend_size/total_bsend_calls;
  if (total_isend_calls > 0)
      isend_avg_size = total_isend_size/total_isend_calls;
  if (total_col_send_calls > 0)
      col_send_avg_size = total_col_send_size/total_col_send_calls;

#if defined(ENABLE_STDDEV)
  /* Now to calculate the standard deviation */
  for (i=0;i<NBINS;i++) {
    tempstore = (send_data.b[i].size + sendrecv_data.b[i].size + 
		 sendrecv_replace_data.b[i].size + 
              bsend_data.b[i].size+ rsend_data.b[i].size + 
		 ssend_data.b[i].size - bsend_avg_size);
    
    bsend_stddev_size += tempstore*tempstore;

    tempstore = (ibsend_data.b[i].size + isend_data.b[i].size + 
		 issend_data.b[i].size
      + irsend_data.b[i].size + bcast_data.b[i].size + gather_data.b[i].size +
		 gatherv_data.b[i].size
      + reduce_data.b[i].size + reduce_scatter_data.b[i].size + 
		 scan_data.b[i].size
      + scatter_data.b[i].size + scatterv_data.b[i].size - isend_avg_size);

    isend_stddev_size += tempstore*tempstore;
    
    tempstore = (allgather_data.b[i].size + allgatherv_data.b[i].size + 
		 allreduce_data.b[i].size 
	       + alltoall_data.b[i].size + alltoallv_data.b[i].size - 
		 col_send_avg_size);
    
    col_send_stddev_size += tempstore*tempstore;
  }  

  if (total_bsend_calls > 0) 
      bsend_stddev_size = sqrt( bsend_stddev_size/total_bsend_calls );
  if (total_isend_calls > 0) 
      isend_stddev_size = sqrt( isend_stddev_size/total_isend_calls );
  if (total_col_send_calls > 0) 
      col_send_stddev_size = sqrt( col_send_stddev_size/total_col_send_calls );
#endif
  
  for (i=0;i<MAX_PROCS;i++) {
    /* Blocking sends - detailed (for nearest neighbor info) */
    total_bsend_comm_partners[i] = send_data.p[i].partners + 
      sendrecv_data.p[i].partners + sendrecv_replace_data.p[i].partners + 
      bsend_data.p[i].partners + rsend_data.p[i].partners + 
	ssend_data.p[i].partners;

    total_bsend_comm_sizes[i] = send_data.p[i].sizes + 
	sendrecv_data.p[i].sizes 
      + sendrecv_replace_data.p[i].sizes  + bsend_data.p[i].sizes
      + rsend_data.p[i].sizes + ssend_data.p[i].sizes;
    
    total_bsend_comm_times[i] = send_data.p[i].times + 
	sendrecv_data.p[i].times 
      + sendrecv_replace_data.p[i].times  + bsend_data.p[i].times
      + rsend_data.p[i].times + ssend_data.p[i].times;

    /* Immediate sends - detailed */
    total_isend_comm_partners[i] = ibsend_data.p[i].partners + 
      isend_data.p[i].partners + issend_data.p[i].partners +
      irsend_data.p[i].partners + bcast_data.p[i].partners + 
      gather_data.p[i].partners + gatherv_data.p[i].partners +
      reduce_data.p[i].partners + reduce_scatter_data.p[i].partners + 
      scan_data.p[i].partners + scatter_data.p[i].partners + 
      scatterv_data.p[i].partners;

    total_isend_comm_sizes[i] = ibsend_data.p[i].sizes + 
	isend_data.p[i].sizes +
      issend_data.p[i].sizes + irsend_data.p[i].sizes + 
	bcast_data.p[i].sizes + 
      gather_data.p[i].sizes + gatherv_data.p[i].sizes + 
	reduce_data.p[i].sizes + 
      reduce_scatter_data.p[i].sizes + scan_data.p[i].sizes +
      scatter_data.p[i].sizes + scatterv_data.p[i].sizes;

    total_isend_comm_times[i] = ibsend_data.p[i].times + 
	isend_data.p[i].times + 
      issend_data.p[i].times + irsend_data.p[i].times + 
	bcast_data.p[i].times + 
      gather_data.p[i].times + gatherv_data.p[i].times+ 
	reduce_data.p[i].times + 
      reduce_scatter_data.p[i].times + scan_data.p[i].times +
      scatter_data.p[i].times + scatterv_data.p[i].times;

    /* Collective - detailed */
    total_col_send_comm_partners[i] = allgather_data.p[i].partners + 
      allgatherv_data.p[i].partners + allreduce_data.p[i].partners + 
      alltoall_data.p[i].partners + alltoallv_data.p[i].partners;
    
    total_col_send_comm_sizes[i] = allgather_data.p[i].sizes + 
      allgatherv_data.p[i].sizes + allreduce_data.p[i].sizes + 
      alltoall_data.p[i].sizes + alltoallv_data.p[i].sizes;
    
    total_col_send_comm_times[i] = allgather_data.p[i].times + 
      allgatherv_data.p[i].times + allreduce_data.p[i].times + 
      alltoall_data.p[i].times + alltoallv_data.p[i].times;
    
  }
  /* End Experimental */

  /* Point-to-point sends */

  /* Blocking */
  getrange_int_single(total_bsend_calls, &max_total_bsend_calls, 
		      &max_loc, &min_total_bsend_calls, 
		      &min_loc, &sum_total_bsend_calls);

  getrange_msgsize_single(total_bsend_size, &max_total_bsend_size, 
		       &max_loc, &min_total_bsend_size, 
		       &min_loc, &sum_total_bsend_size);

  getrange_double_single(total_bsend_time, &max_total_bsend_time, 
		       &max_loc, &min_total_bsend_time, 
		       &min_loc, &sum_total_bsend_time);
  
  if (total_bsend_time > 0) 
      comm_bsend_rate = (double) total_bsend_size/total_bsend_time;

  getrange_double_single(comm_bsend_rate, &max_comm_bsend_rate, 
			 &max_loc, &min_comm_bsend_rate, &min_loc,
			 &sum_comm_bsend_rate);

  getrange_int_single(total_isend_calls, &max_total_isend_calls, 
		      &max_loc, &min_total_isend_calls, 
		      &min_loc, &sum_total_isend_calls);

  getrange_msgsize_single(total_isend_size, &max_total_isend_size, 
		       &max_loc, &min_total_isend_size, 
		       &min_loc, &sum_total_isend_size);

  getrange_double_single(total_isend_time, &max_total_isend_time, 
		       &max_loc, &min_total_isend_time, 
		       &min_loc, &sum_total_isend_time);
  
  if ( rank == 0 && sum_total_isend_calls > 0 && total_isend_time > 0) 
      comm_isend_rate = (double) total_isend_size/total_isend_time;
  
  getrange_double_single(comm_isend_rate, &max_comm_isend_rate, 
			 &max_loc, &min_comm_isend_rate, &min_loc,
			 &sum_comm_isend_rate);

  /* Collective sends */
  getrange_int_single(total_col_send_calls, &max_total_col_send_calls, 
		      &max_loc, &min_total_col_send_calls, 
		      &min_loc, &sum_total_col_send_calls);

  getrange_msgsize_single(total_col_send_size, &max_total_col_send_size, 
		       &max_loc, &min_total_col_send_size, 
		       &min_loc, &sum_total_col_send_size);

  getrange_double_single(total_col_send_time, &max_total_col_send_time, 
			 &max_loc, &min_total_col_send_time, 
			 &min_loc, &sum_total_col_send_time);

  if (total_col_send_time > 0) 
      comm_col_rate	= (double) total_col_send_size/total_col_send_time;

  getrange_double_single(comm_col_rate, &max_comm_col_rate, &max_loc,
			 &min_comm_col_rate, &min_loc,
			 &sum_comm_col_rate);

  /* Sync time - time spent waiting for other procs */
  getrange_double_single(total_col_timesync, &max_total_col_timesync, 
			 &max_loc, &min_total_col_timesync, 
			 &min_loc, &sum_total_col_timesync);
  
  if ( rank == 0 && sum_total_col_send_calls > 0 ) {
      double temptime;
    fprintf(pf,"\n");
    fprintf(pf,"Communication Time:");
    temptime = (sum_total_bsend_time + sum_total_isend_time + 
		 sum_total_col_send_time + sum_total_col_timesync) / nprocs;
    fprintf(pf,"\t%.4f sec", temptime);

    fprintf(pf,"\n");
    fprintf(pf,"Communication Volume:");
    tempstore = (sum_total_bsend_size + 
         sum_total_isend_size + sum_total_col_send_size) / 
         ((double) nprocs * ONE_MB);
#if SIZEOF_LONG_LONG >= 8
    fprintf(pf,"\t%lld MB\n", tempstore);
#else
    fprintf(pf,"\t%.4f MB\n", tempstore);
#endif
  }
}

#ifdef UNUSED
static void printout_header( FILE *pf, const char *name, int llrank )
{
  if ( llrank == 0 ){
      fprintf(pf, "\n------   %-40s ------\n", name);
  }
}
#endif



/* For each rank, calculate the total number of communication partners */
/* Print the average across ranks.                                     */
static void printout_comm_dests( int32_t target[], const char *name,
				 void *handle )
{
#ifdef GTABLE
/*     Gtable *gtable = (Gtable *)handle; */
#else
    FILE *pf = (FILE *pf) handle;
#endif    
  int i,j;
  int32_t SelectBit;

  int nprocs, rank;
  int partners, max_partners, min_partners, sum_partners;
  int max_loc, min_loc;

  PMPI_Comm_size( MPI_COMM_WORLD, &nprocs );
  PMPI_Comm_rank( MPI_COMM_WORLD, &rank );

  partners = 0;
  for (i=0;i<MAX_PROCBY32;i++) {
    SelectBit = 1;
    for (j=0;j<32;j++) {
      if (Bit_Is_Set(target,SelectBit)) partners+=1;
      SelectBit = SelectBit << 1;
    }
  }

  /* Get the min/max/avg across procs, also get min_loc/max_loc */
  getrange_int_single(partners, &max_partners, 
		      &max_loc, &min_partners, 
		      &min_loc, &sum_partners);

  if ( rank == 0 && sum_partners > 0 ) {
/*#ifdef GTABLE
    GtableNewRow( gtable );
    GtableAddEntryFmt( gtable, -1, -1, "%s", name);
    GtableAddEntryFmt( gtable, -1, -1, "Max partners: %d", max_partners);
#else
    fprintf( pf, "%-20s\t", name );
    fprintf( pf, "%d\n", max_partners );
    #endif */ /* GTABLE */
  }
}


void  OutputCommunicationData( FILE *pf, int llrank, int nprocs )
{  
    void *handle = pf;
/*  if (llrank == 0)
      PRINT_HEADER("Barriers and Waits");*/

  /*  if ( llrank == 0 ){
    GtableNewRow( gtable );
    GtableAddEntry( gtable, -1, -1, "[max bin limit]  [min/max/avg calls] [min rank/max rank]  [min/max/avg msg size] [min rank/max rank] [min/max/avg time] [min rank/max rank]");
	  }*/


#define PRINT_MSG_STATS(sname,name) \
printout_msgstats_coll( pf, &sname##_data, 0, name, llrank, nprocs )
#define PRINT_MSG_STATS_COLL(sname,name) \
printout_msgstats_coll( pf, &sname##_data, sname##_timesync, name, llrank, nprocs )

  if (FullOutput) {
      if (llrank == 0) {
	  /* Add column labels */
	  fprintf( pf, "%32s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%32s\t%32s\n", 
		   "Routine", "Calls", "ave time", "min time", "max time",
		   "ave len", "min len", "max len", "ave sync", "min sync", 
                   "max sync", "Calls by message size", 
		   "% data by message size");
      }
      PRINT_MSG_STATS_COLL( allgather, "MPI_Allgather" );
      PRINT_MSG_STATS_COLL( allgatherv, "MPI_Allgatherv" );
      PRINT_MSG_STATS_COLL( allreduce, "MPI_Allreduce" );
      PRINT_MSG_STATS_COLL( alltoall, "MPI_Alltoall" );
      PRINT_MSG_STATS_COLL( alltoallv, "MPI_Alltoallv" );
      if (llrank == 0) {
          fprintf( pf, "\n");
	  /* Add column labels */
	  fprintf( pf, "%32s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%32s\t%32s\n", 
		   "Routine", "Calls", "ave time", "min time", "max time",
		   "ave len", "min len", "max len", "Calls by message size", 
		   "% data by message size");
      }
      PRINT_MSG_STATS( bcast, "MPI_Bcast" );
      PRINT_MSG_STATS( gather, "MPI_Gather" );
      PRINT_MSG_STATS( gatherv, "MPI_Gatherv" );
      PRINT_MSG_STATS( reduce_scatter, "MPI_Reduce_scatter" );
      PRINT_MSG_STATS( reduce, "MPI_Reduce" );
      PRINT_MSG_STATS( scan, "MPI_Scan" );
      PRINT_MSG_STATS( scatter, "MPI_Scatter" );
      PRINT_MSG_STATS( scatterv, "MPI_Scatterv" );
      
      PRINT_MSG_STATS( send, "MPI_Send" );
      PRINT_MSG_STATS( sendrecv, "MPI_Sendrecv" );
      PRINT_MSG_STATS( sendrecv_replace, "MPI_Sendrecv_replace" );
      PRINT_MSG_STATS( bsend, "MPI_Bsend" );
      PRINT_MSG_STATS( ibsend, "MPI_Ibsend" );
      PRINT_MSG_STATS( isend, "MPI_Isend" );
      PRINT_MSG_STATS( issend, "MPI_Issend" );
      PRINT_MSG_STATS( irsend, "MPI_Irsend" );
      PRINT_MSG_STATS( rsend, "MPI_Rsend" );
      PRINT_MSG_STATS( ssend, "MPI_Ssend" );
      PRINT_MSG_STATS( recv, "MPI_Recv" );
      PRINT_MSG_STATS( irecv, "MPI_Irecv" );
      PRINT_MSG_STATS( pack, "MPI_Pack" );
      PRINT_MSG_STATS( unpack, "MPI_Unpack" );
      
      /*   PRINT_HEADER("Message Totals");*/
      printout_total_msgstats(handle, llrank, nprocs);
  }
}
/*
 * ----------------------------------------------------------------------
 * The following code provides access to hardware counters
 * ----------------------------------------------------------------------
 */
/* ----------------------------------------------------------------------
 * Output Usage Information
 * ------------------------------------------------------------------------ */
#ifdef HAVE_PROCSTAT
int ReadProcStat( procstat_t *);
#endif

void OutputResourceUsage( FILE *pf, int llrank, int nprocs )
{
  int max_wall_clock_rank, min_wall_clock_rank;
  double wall_clock_t;
  double max_wall_clock_t, min_wall_clock_t;
  double t_tmp;

#ifdef HAVE_GETRUSAGE
  struct rusage myusage;
  double sys_t=0.0;
  double max_sys_t, min_sys_t;
  int max_sys_rank, min_sys_rank;
  double user_t=0.0;
  double max_user_t, min_user_t;
  int max_user_rank, min_user_rank;
  long maxrss, idrss;
  long minflt, majflt;
  long nswap; 
  long inblock, oublock;
  long msgsnd, msgrcv; 
  long nsignals; 
  long nvcsw, nivcsw;
#ifdef GETRUSAGE_HAS_MAXRSS
    long maxrss_all, minrss_all, sumrss_all, rss_tmp;
#endif /* GETRUSAGE_HAS_MAXRSS */
#endif /* HAVE_GETRUSAGE */
#ifdef HAVE_PROCSTAT
  int ierr;
  procstat_t stats;
#ifndef HAVE_GETRUSAGE
  long maxrss;
#ifndef GETRUSAGE_HAS_MAXRSS
  long maxrss_all, minrss_all;
#endif /* NOT GETRUSAGE_HAS_MAXRSS */
#endif /* NOT HAVE_GETRUSAGE */
#endif /* HAVE_PROCSTAT */

  MPI_End_time = clock();
  wall_clock_t = ((double) MPI_End_time - MPI_Start_time)/CLOCKS_PER_SEC;
  t_tmp = wall_clock_t;
  getrange_double( &t_tmp, 1, &wall_clock_t, &min_wall_clock_t, 
		   &max_wall_clock_t, &min_wall_clock_rank, 
		   &max_wall_clock_rank );
  wall_clock_t /= nprocs;

  if (llrank == 0) {
      switch (outform) {
      case OUT_TEXT:
      case OUT_LATEX:
	  fprintf( pf, "Execute time:\t%.4g\n", wall_clock_t );
	  break;
      case OUT_XML:
	  fprintf(pf,"<exectime>%.4g</exectime>\n", wall_clock_t);
	  break;
      }
      fprintf(pf, "Timing Stats: [seconds]\t[min/max]    \t[min rank/max rank]\n");
      fprintf(pf, "  wall-clock: %.4g sec\t%f / %f\t%d / %d\n",
              wall_clock_t, min_wall_clock_t, max_wall_clock_t,
              min_wall_clock_rank, max_wall_clock_rank);
  }

#ifdef HAVE_GETRUSAGE
  /* Get Resource usage */
  if ( getrusage(RUSAGE_SELF, &myusage) < 0 )
    fprintf(stderr, "[%d] getrusage error\n", llrank);
  user_t   = (double) myusage.ru_utime.tv_sec +
              myusage.ru_utime.tv_usec / 1000000.0;
  t_tmp = user_t;
  getrange_double( &t_tmp, 1, &user_t, &min_user_t, &max_user_t, 
		   &min_user_rank, &max_user_rank );
  user_t       /= nprocs;
  sys_t	   = (double) myusage.ru_stime.tv_sec +
              myusage.ru_stime.tv_usec / 1000000.0;
  t_tmp = sys_t;
  getrange_double( &t_tmp, 1, &sys_t, &min_sys_t, &max_sys_t, 
		   &min_sys_rank, &max_sys_rank );
  sys_t	       /= nprocs;

  if (llrank == 0) {
    fprintf(pf, "        user: %.4g sec\t%f / %f\t%d / %d\n",
	    user_t, min_user_t, max_user_t, min_user_rank, max_user_rank);
    fprintf(pf, "         sys: %.4g sec\t%f / %f\t%d / %d\n",
	    sys_t, min_sys_t, max_sys_t, min_sys_rank, max_sys_rank);
  }
  maxrss   = myusage.ru_maxrss;
  idrss	   = myusage.ru_idrss;
  minflt   = myusage.ru_minflt;
  majflt   = myusage.ru_majflt;
  nswap	   = myusage.ru_nswap;
  inblock  = myusage.ru_inblock;
  oublock  = myusage.ru_oublock;
  msgsnd   = myusage.ru_msgsnd;
  msgrcv   = myusage.ru_msgrcv;
  nsignals = myusage.ru_nsignals;
  nvcsw	   = myusage.ru_nvcsw;
  nivcsw   = myusage.ru_nivcsw;

#ifdef GETRUSAGE_HAS_MAXRSS
  PMPI_Reduce( &maxrss, &maxrss_all, 1, MPI_LONG, MPI_MAX, 0, MPI_COMM_WORLD);
  PMPI_Reduce( &maxrss, &minrss_all, 1, MPI_LONG, MPI_MIN, 0, MPI_COMM_WORLD);
  if (llrank == 0) {
    fprintf(pf, "\n");
    fprintf(pf, "Memory Usage Stats (RSS) [min/max KB]:\t%ld/%ld\n",
	  minrss_all, maxrss_all);
    fprintf(pf, "Memory Usage Stats (RSS):\t%ld KB\n", maxrss_all);
  }  
#endif /* GETRUSAGE_HAS_MAXRSS */
#endif /* HAVE_GETRUSAGE */

#ifdef HAVE_PROCSTAT
  ierr = ReadProcStat( &stats );
  maxrss = stats.rss;
  PMPI_Reduce( &maxrss, &maxrss_all, 1, MPI_LONG, MPI_MAX, 0, MPI_COMM_WORLD);
  PMPI_Reduce( &maxrss, &minrss_all, 1, MPI_LONG, MPI_MIN, 0, MPI_COMM_WORLD);
  if (llrank == 0) {
    fprintf(pf, "Memory Usage Stats (RSS) [min/max KB]:\t%ld/%ld\n",
       minrss_all, maxrss_all);
    fprintf(pf, "Memory Usage Stats (RSS) :\t%ld KB\n",maxrss_all);
  }
#endif /* HAVE_PROCSTAT */

  if (llrank == 0) {
    fprintf( pf, "\n" );
  }

}

void OutputHeader( FILE *pf, int llrank )
{
  if ( llrank == 0 ) {
    switch (outform) {
    case OUT_XML:
      fprintf(pf,"<?xml version=\"1.0\" ?>\n");
      fprintf(pf,"<?xml-stylesheet type=\"text/xsl\" href=\"xml2html.xsl\"?>\n");
      fprintf(pf,"<FPMPI_OUTPUT>\n");
      break;
    case OUT_TEXT:
      fprintf(pf, "MPI Routine Statistics\n" );
      /* Set the environment variable FPMPI_NO_HEADER to suppress this
	 header message */
      if (!getenv("FPMPI_NO_HEADER")) {
	fprintf( pf, "Explanation of data:\n\
Times are the time to perform the operation, e.g., the time for MPI_Send\n\
Average times are the average over all processes, e.g., sum (time on each\n\
process) / number of processes\n\
Min and max values are over all processes\n\
(Data is always average/min/max)\n\
Amount of data is computed in bytes.  For point-to-point operations,\n\
it is the data sent or received.  For collective operations, it is the\n\
data contibuted to the operation.  E.g., for an MPI_Bcast, the amount of\n\
data is the number of bytes provided by the root, counted only at the root.\n\
For synchronizing collective operations, the average, min, and max time\n\
spent synchronizing is shown next.\n\
Calls by message size shows the fraction of calls that sent messages of a \n\
particular size.  The bins are\n\
0 bytes, 1-4 bytes, 5-8 bytes, 9-16, 17-32, 33-64, -128, -256, -512, -1024\n\
 -4K, -8K, -16K, -32K, -64K, -128K, -256K, -512K, -1M, -4M, -8M, -16M, \n\
 -32M, -64M, -128M, -256M, -512M, -1GB, >1GB.\n\
Each bin is represented by a single digit, representing the 10's of percent\n\
of messages within this bin.  A 0 represents precisely 0, a . (period) \n\
represents more than 0 but less than 10%%.  A * represents 100%%.\n\
Messages by message size shows similar information, but for the total\n\
message size.\n\
\n\
The experimental topology information shows the 1-norm distance that the\n\
longest point-to-point message travelled, by process.\n\
\n\
MPI_Pcontrol may be used to control the collection of data\n" );
      }

      break;
    case OUT_LATEX:
      fprintf( pf, "\\documentclass{article}\n\\begin{document}\n" );
      break;
    }
  }
}

void  OutputTail( FILE *pf, int llrank )
{
  if ( llrank == 0 ) {
    switch (outform) {
    case OUT_XML:
      fprintf(pf,"</FPMPI_OUTPUT>\n");
      break;
    case OUT_TEXT:
      break;
    case OUT_LATEX:
      fprintf( pf, "\\end{document}\n" );
      break;
    }
    fclose(pf);
  }
}

void  OutputProgramDescription( FILE *pf, 
				int llrank, int nprocs )
{
    if ( llrank == 0 ) {
	if (cmdname) {
	  switch (outform) {
	  case OUT_XML:
	    fprintf( pf, "<cmdname>%s</cmdname>\n", cmdname );
	    break;
	  case OUT_TEXT:
	    fprintf( pf, "Command: %s\n", cmdname );
	    break;
	  case OUT_LATEX:
	    fprintf( pf, "Command: %s\n", cmdname );
	    break;
	  }
	}

	{
	    time_t t;
	    char *c;
	    int  clen;
	    t = time(NULL);
	    c = asctime(localtime(&t));
	    /* Remove any trailing newline on c */
	    clen = strlen( c );
	    if (c[clen-1] == '\n') c[clen-1] = 0;
	    fprintf(pf,"\n");
	    switch (outform) {
	    case OUT_XML:
	      fprintf(pf,"<timestamp> %s</timestamp>\n", c );
	      fprintf(pf,"<numprocs>%d</numprocs>\n", nprocs);
	      break;
	    case OUT_TEXT:
	    case OUT_LATEX:
	      fprintf(pf,"Date:\t%s\n", c );
	      fprintf(pf,"Processes:\t%d\n", nprocs);
	      break;
	    }
	}
	
    }
    
}

#ifndef MAXPATHLEN
#define MAXPATHLEN 1024
#endif
/* 
 * The output filename can be controlled with environment variables:
 *
 */
FILE *OutputOpenFile( int llrank )
{
    FILE *pf = 0;
    char profname[MAXPATHLEN];
    char *ext="txt";
    const char *pname;

    if ( llrank == 0 ) {
	pname = getenv("MPI_PROFILE_FILE_NAME");
	if (!pname) {
	    switch(outform) {
	    case OUT_XML: ext = "xml"; break;
	    case OUT_TEXT: ext = "txt"; break;
	    case OUT_LATEX: ext = "tex"; break;
	    }
	    sprintf(profname,"%s.%s", MPI_PROF_FILE_NAME,ext);
	    if (getenv("MPI_PROFILE_DIR") != NULL)
		sprintf(profname, "%s/fpmpi_output.%s", 
			getenv("MPI_PROFILE_DIR"),ext);
	    pname = (const char *)profname;
	}
	
	if (Verbose) 
	    printf("Profile will be saved in %s.\n", pname);
	pf = fopen( pname, "w" );
	if (!pf) {
	    fprintf( stderr, "Could not open %s\n", pname );
	    perror( "Reason: " );
	}
    }
    return pf;
}

void  OutputPartners( FILE *pf, int llrank ) 
{
    int32_t   alltarget[MAX_PROCBY32];
    int       i, j, n, nprocs;
    msgsize_t alldest[MAX_PROCS][NBINS];
#ifdef GTABLE
#define PRINT_COMM_DESTS(sname,name) \
  printout_comm_dests( sname##_target, name, pf/*gtable*/ )
#else
#define PRINT_COMM_DESTS(sname,name) \
  printout_comm_dests( sname##_target, name, pf )
#endif /* GTABLE */

  if (FullOutput) {
      /*  PRINT_HEADER("Communication Partners");*/
      PRINT_COMM_DESTS( send, "MPI_Send" );
      PRINT_COMM_DESTS( sendrecv, "MPI_Sendrecv" );
      PRINT_COMM_DESTS( sendrecv_replace, "MPI_Sendrecv_replace" );
      PRINT_COMM_DESTS( bsend, "MPI_Bsend" );
      PRINT_COMM_DESTS( ibsend, "MPI_Ibsend" );
      PRINT_COMM_DESTS( isend, "MPI_Isend" );
      PRINT_COMM_DESTS( issend, "MPI_Issend" );
      PRINT_COMM_DESTS( irsend, "MPI_Irsend" );
      PRINT_COMM_DESTS( rsend, "MPI_Rsend" );
      PRINT_COMM_DESTS( ssend, "MPI_Ssend" );
      PRINT_COMM_DESTS( recv, "MPI_Recv" );
      PRINT_COMM_DESTS( irecv, "MPI_Irecv" );
  }

    /* For each communication type, gather up the partner list */
    MPI_Comm_size( MPI_COMM_WORLD, &nprocs );
    n = (nprocs + 31) / 32;
    for (i=0; i<n; i++) {
	alltarget[i] = recv_target[i] | send_target[i] | sendrecv_target[i] |
	    sendrecv_replace_target[i] | bsend_target[i] | ibsend_target[i] |
	    isend_target[i] | issend_target[i] | irsend_target[i] |
	    rsend_target[i] | ssend_target[i];
	for (j=0; j<NBINS; j++) {
	    alldest[i][j] = recv_targetsize[i][j] + send_targetsize[i][j] + 
		sendrecv_targetsize[i][j] +
	    sendrecv_replace_targetsize[i][j] + bsend_targetsize[i][j] + 
		ibsend_targetsize[i][j] +
	    isend_targetsize[i][j] + issend_targetsize[i][j] + 
		irsend_targetsize[i][j] +
	    rsend_targetsize[i][j] + ssend_targetsize[i][j];
	}
    }
    fpmpiOutputPartnerTopology( alltarget, pf, llrank );
    fpmpiOutputDetailedPartners( alltarget, alldest, pf, llrank );
}
